import asyncio
import logging
import ssl

#

L = logging.getLogger(__name__)

#


class Stream(object):
    """
    Description: This object represent a client connection.
    It is unencrypted STREAM socket.
    """

    def __init__(self, loop, socket, outbound_queue=None):
        self.Loop = loop
        self.Socket = socket
        if outbound_queue is None:
            self.OutboundQueue = asyncio.Queue()
        else:
            self.OutboundQueue = outbound_queue

    async def recv_into(self, buf):
        return await self.Loop.sock_recv_into(self.Socket, buf)

    def send(self, data):
        self.OutboundQueue.put_nowait(data)

    async def outbound(self):
        """
        Handle outbound direction
        """
        while True:
            data = await self.OutboundQueue.get()
            if data is None:
                break
            await self.Loop.sock_sendall(self.Socket, data)

    async def close(self):
        self.Socket.close()


class TLSStream(object):
    """
    Description: This object represent a TLS client connection.
    It is encrypted SSL/TLS socket abstraction.
    """

    def __init__(self, loop, sslcontext, socket, server_side: bool):
        """
        Description:


        """
        self.Loop = loop
        self.Socket = socket

        self.SSLContext = sslcontext

        self.InBuffer = ssl.MemoryBIO()
        self.OutBuffer = ssl.MemoryBIO()
        self.OutboundQueue = asyncio.Queue()

        self.SSLSocket = sslcontext.wrap_bio(
            self.InBuffer, self.OutBuffer, server_side=server_side
        )

    async def handshake(self):
        """
        Description:

        :return: False if error is raised or socket is closed, otherwise returns True.

        |

        """
        while True:
            try:
                self.SSLSocket.do_handshake()
                break

            except ssl.SSLWantReadError:
                if self.OutBuffer.pending > 0:
                    data = self.OutBuffer.read()
                    await self.Loop.sock_sendall(self.Socket, data)

                data = await self.Loop.sock_recv(self.Socket, 4096)
                if len(data) == 0:
                    # Socket has been closed
                    # TODO: self.Socket.shutdown()
                    return False

                self.InBuffer.write(data)
                continue

            except Exception:
                L.exception("SSL handshake failed")
                self.Socket.close()
                return False

        return True

    async def recv_into(self, buf):
        """
        Description:


        """
        while True:
            try:
                data = self.SSLSocket.read(len(buf))
                if len(data) == 0:
                    # Socket has been closed
                    return 0

                buf[: len(data)] = data
                return len(data)

            except ssl.SSLWantReadError:
                if self.OutBuffer.pending > 0:
                    data = self.OutBuffer.read()
                    await self.Loop.sock_sendall(self.Socket, data)

                data = await self.Loop.sock_recv(self.Socket, 4096)
                if len(data) == 0:
                    # Socket has been closed
                    # TODO: self.Socket.shutdown()
                    return 0

                self.InBuffer.write(data)
                continue

    def send(self, data):
        """
        Description:


        """
        self.OutboundQueue.put_nowait(data)

    async def outbound(self):
        """
        Handle outbound direction
        """
        while True:
            data = await self.OutboundQueue.get()
            if data is None:
                break

            self.SSLSocket.write(data)

            while self.OutBuffer.pending > 0:
                data = self.OutBuffer.read()
                await self.Loop.sock_sendall(self.Socket, data)

    async def close(self):
        """
        Description:


        """
        self.OutboundQueue.put_nowait(None)
        self.Socket.close()



from .trigger import Trigger

###


class RunOnceTrigger(Trigger):
    """
    RunOnceTrigger is an obsolete trigger that should not be used.

    The following issues should be taken into consideration and need to be refactored in the future:
    1) RunOnceTrigger issues the application stop.
    2) The self.fire() event is triggered before the pipeline is_ready()
    """

    def __init__(self, app, id=None):
        super().__init__(app, id)
        self.App = app
        self.Loop.call_soon(self.fire)

    def done(self, trigger_source):
        self.App.stop()



import aiohttp
import logging
import asyncio
import json
from ..abc.connection import Connection


L = logging.getLogger(__name__)


class SlackConnection(Connection):
    ConfigDefaults = {
        "hook_url": "https://hooks.slack.com/services/T00000000/B00000000/XXXXXXXXXXXXXXXXXXXXXXXX",
        "output_queue_max_size": 10,
    }

    def __init__(self, app, id=None, config=None):
        super().__init__(app, id=id, config=config)

        self.Url = self.Config["hook_url"]

        self.Loop = app.Loop
        self.PubSub = app.PubSub

        self._output_queue_max_size = int(self.Config["output_queue_max_size"])
        self._output_queue = asyncio.Queue(maxsize=self._output_queue_max_size + 1)
        self.LoaderTask = asyncio.ensure_future(self._loader(), loop=self.Loop)

        self.PubSub.subscribe("Application.exit!", self._on_exit)

    async def _on_exit(self, event_name):
        # Wait till the _loader() terminates
        pending = [self.LoaderTask]
        while len(pending) > 0:
            # By sending None via queue, we signalize end of life
            await self._output_queue.put(None)
            done, pending = await asyncio.wait(
                pending, return_when=asyncio.FIRST_COMPLETED
            )

    def consume(self, slack_message):
        slack_message = json.dumps(slack_message)
        self._output_queue.put_nowait(slack_message)

        if self._output_queue.qsize() == self._output_queue_max_size:
            self.PubSub.publish("SlackConnection.pause!", self)

    async def _loader(self):
        async with aiohttp.ClientSession() as session:
            while True:
                slack_message = await self._output_queue.get()
                if slack_message is None:
                    break

                if self._output_queue.qsize() == self._output_queue_max_size - 1:
                    self.PubSub.publish(
                        "SlackConnection.unpause!", self, asynchronously=True
                    )

                async with session.post(
                    self.Url,
                    data=slack_message,
                    headers={"content-type": "application/json"},
                ) as resp:
                    resp_body = await resp.text()
                    if resp.status != 200:
                        L.error(
                            "Failed to send Slack message status:{} body:{}".format(
                                resp.status, resp_body
                            )
                        )

                        # Requeue the message, wait a bit and try again
                        self._output_queue.put_nowait(slack_message)
                        await asyncio.sleep(5)



import pprint
import logging

from ...abc import Expression, evaluate

###

L = logging.getLogger("bspump.declarative.DEBUG")

###


class DEBUG(Expression):
    Attributes = {
        "What": ["*"],  # TODO: This ...
    }

    def __init__(self, app, *, arg_what):
        super().__init__(app)
        self.What = arg_what

    def __call__(self, context, event, *args, **kwargs):
        ret = evaluate(self.What, context, event, *args, **kwargs)
        L.warning(pprint.pformat(ret))
        return ret



import binascii

from ..abc.processor import Processor


class HexlifyProcessor(Processor):
    """
    Description:

    |

    """

    def process(self, context, event):
        """
        Description:

        :return: binascii.hexlify(event)

        |

        """
        return binascii.hexlify(event)



import logging
import json

from bspump.asab import Configurable

###

L = logging.getLogger(__name__)

###


class Model(Configurable):
    """
    Generic `Model` object. Loads trained model and parameters.

    """

    ConfigDefaults = {
        "path_model": "",  # path to serialized model
        "path_parameters": "",  # path to serialized model
    }

    def __init__(self, app, id=None, config=None):
        self.Id = id if id is not None else self.__class__.__name__
        super().__init__("model:{}".format(self.Id), config=config)
        self.PathModel = self.Config["path_model"]
        self.PathParameters = self.Config["path_parameters"]
        self.App = app
        self.Loop = app.Loop

    def load_model_from_file(self):
        """
        Load model from file.
        """
        raise NotImplementedError()

    def load_parameters_from_file(self):
        """
        Loads model parameters from json file. Override if needed.
        """
        with open(self.PathParameters) as f:
            self.Parameters = json.load(f)

    async def update(self):
        """
        Updates model on fly.
        """
        pass

    def transform(self, *args):
        """
        Method used to transform data for model input.
        """
        raise NotImplementedError()

    def predict(self, *args):
        """
        Method uses model to predict value from sample.
        """
        raise NotImplementedError()



from .test_bytes import *
from .test_flatten import *
from .test_hexlify import *
from .test_iterator import *
from .test_json import *

# from .test_jsonbytes import *
from .test_mapping import *
from .test_null import *
from .test_print import *

# TODO test_routing
# TODO test_tee
from .test_time import *



import asyncio
import logging
import datetime

import aiohttp

from ..web.rest.json import json_response
from ..log import LOG_NOTICE
from ..web.auth import noauth

##

L = logging.getLogger(__name__)

##


class WebApiLoggingHandler(logging.Handler):
    def __init__(self, app, level=logging.NOTSET, buffer_size: int = 10):
        super().__init__(level=level)

        self.Buffer = []
        self._buffer_size = buffer_size
        self.WebSockets = set()

        app.PubSub.subscribe("Application.stop!", self._on_stop)

    async def _on_stop(self, _on_stop, x):
        for ws in list(self.WebSockets):
            await ws.send_json(
                {
                    "t": datetime.datetime.utcnow().isoformat()
                    + "Z",  # This is OK, no tzinfo needed
                    "C": "asab.web",
                    "M": "Closed.",
                    "l": logging.INFO,
                }
            )
            await ws.close()

    def emit(self, record):
        if logging.DEBUG < record.levelno <= logging.INFO:
            severity = 6  # Informational
        elif record.levelno <= LOG_NOTICE:
            severity = 5  # Notice
        elif record.levelno <= logging.WARNING:
            severity = 4  # Warning
        elif record.levelno <= logging.ERROR:
            severity = 3  # Error
        elif record.levelno <= logging.CRITICAL:
            severity = 2  # Critical
        else:
            severity = 1  # Alert

        log_entry = {
            "t": datetime.datetime.utcfromtimestamp(record.created).isoformat() + "Z",
            "C": record.name,
            "s": "{}:{}".format(record.funcName, record.lineno),
            "p": record.process,
            "Th": record.thread,
            "l": severity,
        }

        message = record.getMessage()
        if record.exc_text is not None:
            message += "\n" + record.exc_text
        if record.stack_info is not None:
            message += "\n" + record.stack_info
        if len(message) > 0:
            log_entry["M"] = message

        sd = record.__dict__.get("_struct_data")
        if sd is not None:
            log_entry["sd"] = sd

        if len(self.Buffer) > self._buffer_size:
            del self.Buffer[0]
            self.Buffer.append(log_entry)

        else:
            self.Buffer.append(log_entry)

        if len(self.WebSockets) > 0:
            asyncio.ensure_future(self._send_ws(log_entry))

    @noauth
    async def get_logs(self, request):
        """
        Get logs.
        ---
        tags: ['asab.log']
        """

        return json_response(request, self.Buffer)

    @noauth
    async def ws(self, request):
        """
        # Live feed of logs over websocket

        Usable with e.g. with React Lazylog

        ```
        <LazyLog
                url={this.AsabLogWsURL}
                follow
                websocket
                websocketOptions={{
                        formatMessage: e => log_message_format(e),
                }}
        />

        function log_message_format(e) {
                e = JSON.parse(e)
                var msg = e.t;
                if (e.l != undefined) msg += " " + e.l;
                if (e.sd != undefined) msg += ` ${JSON.stringify(e.sd)}`;
                if (e.M != undefined) msg += " " + e.M;
                if (e.C != undefined) msg += ` [${e.C}]`
                return msg;
        }
        ```

        ---
        tags: ['asab.log']
        externalDocs:
                description: React Lazylog
                url: https://github.com/mozilla-frontend-infra/react-lazylog#readme
        """

        ws = aiohttp.web.WebSocketResponse()
        await ws.prepare(request)

        await ws.send_json(
            {
                "t": datetime.datetime.utcnow().isoformat()
                + "Z",  # This is OK, no tzinfo needed
                "C": "asab.web",
                "M": "Connected.",
                "l": logging.INFO,
            }
        )

        # Send historical logs
        for log_entry in self.Buffer:
            await ws.send_json(log_entry)

        self.WebSockets.add(ws)
        try:
            async for msg in ws:
                if msg.type == aiohttp.WSMsgType.ERROR:
                    break

        finally:
            self.WebSockets.remove(ws)

        return ws

    async def _send_ws(self, log_entry):
        for ws in self.WebSockets:
            await ws.send_json(log_entry)



from .aggregator import (
    AggregationStrategy,
    Aggregator,
    ListAggregationStrategy,
    ListEventAggregationStrategy,
    StringAggregationStrategy,
)
from .bytes import BytesToStringParser
from .bytes import StringToBytesParser
from .flatten import FlattenDictProcessor
from .hexlify import HexlifyProcessor
from .iterator import IteratorGenerator
from .iterator import IteratorSource
from .json import StdDictToJsonParser
from .json import StdJsonToDictParser
from .mapping import MappingKeysGenerator, MappingValuesGenerator, MappingItemsGenerator
from .mapping import MappingKeysProcessor, MappingValuesProcessor, MappingItemsProcessor
from .null import NullSink
from .print import (
    PrintSink,
    PPrintSink,
    PrintProcessor,
    PPrintProcessor,
    PrintContextProcessor,
    PPrintContextProcessor,
)
from .routing import DirectSource
from .routing import InternalSource
from .routing import RouterProcessor
from .routing import RouterSink
from .tee import TeeProcessor
from .tee import TeeSource
from .time import TimeZoneNormalizer
from .transfr import MappingTransformator
from .jsonbytes import DictToJsonBytesParser

try:
    from .jq import JQProcessor
except ImportError:
    pass

__all__ = (
    "BytesToStringParser",
    "StringToBytesParser",
    "FlattenDictProcessor",
    "HexlifyProcessor",
    "IteratorGenerator",
    "IteratorSource",
    "JsonToDictParser",
    "StdDictToJsonParser",
    "StdJsonToDictParser",
    "MappingKeysGenerator",
    "MappingValuesGenerator",
    "MappingItemsGenerator",
    "MappingKeysProcessor",
    "MappingValuesProcessor",
    "MappingItemsProcessor",
    "NullSink",
    "PrintSink",
    "PPrintSink",
    "PrintProcessor",
    "PPrintProcessor",
    "PrintContextProcessor",
    "PPrintContextProcessor",
    "InternalSource",
    "RouterProcessor",
    "RouterSink",
    "TeeProcessor",
    "TeeSource",
    "TimeZoneNormalizer",
    "MappingTransformator",
    "DirectSource",
    "AggregationStrategy",
    "Aggregator",
    "ListAggregationStrategy",
    "ListEventAggregationStrategy",
    "StringAggregationStrategy",
    "DictToJsonBytesParser",
    "JQProcessor",
)



import numpy as np
import os


class TimeConfig(object):
    def __init__(self, resolution, columns, start):
        self.DType = [
            ("resolution", "f8"),
            ("columns", "i8"),
            ("start", "f8"),
            ("end", "f8"),
        ]
        self.TC = np.zeros(1, dtype=self.DType)
        self.TC["resolution"][0] = resolution
        self.TC["columns"][0] = columns
        self.TC["start"][0] = start
        self.TC["end"][0] = start - (resolution * columns)

    def get_resolution(self):
        return self.TC["resolution"][0]

    def get_columns(self):
        return self.TC["columns"][0]

    def get_start(self):
        return self.TC["start"][0]

    def get_end(self):
        return self.TC["end"][0]

    def set_resolution(self, resolution):
        self.TC["resolution"][0] = resolution

    def set_start(self, start):
        self.TC["start"][0] = start

    def set_end(self, end):
        self.TC["end"][0] = end

    def add_start(self, time):
        self.TC["start"][0] += time

    def add_end(self, time):
        self.TC["end"][0] += time


class PersistentTimeConfig(TimeConfig):
    def __init__(self, path, resolution, columns, start):
        super().__init__(resolution, columns, start)
        self.Path = path
        if os.path.exists(self.Path):
            self.TC = np.memmap(self.Path, dtype=self.DType, mode="readwrite")
        else:
            tc = np.memmap(self.Path, dtype=self.DType, mode="w+", shape=(1,))
            tc[:] = self.TC[:]
            self.TC = tc



SWAGGER_OAUTH_PAGE = """<!doctype html>
<html lang="en-US">
<head>
	<title>Swagger UI: OAuth2 Redirect</title>
</head>
<body>
<script>
	'use strict';
	function run () {
		var oauth2 = window.opener.swaggerUIRedirectOauth2;
		var sentState = oauth2.state;
		var redirectUrl = oauth2.redirectUrl;
		var isValid, qp, arr;

		if (/code|token|error/.test(window.location.hash)) {
			qp = window.location.hash.substring(1);
		} else {
			qp = location.search.substring(1);
		}

		arr = qp.split("&");
		arr.forEach(function (v,i,_arr) { _arr[i] = '"' + v.replace('=', '":"') + '"';});
		qp = qp ? JSON.parse('{' + arr.join() + '}',
				function (key, value) {
					return key === "" ? value : decodeURIComponent(value);
				}
		) : {};

		isValid = qp.state === sentState;

		if ((
			oauth2.auth.schema.get("flow") === "accessCode" ||
			oauth2.auth.schema.get("flow") === "authorizationCode" ||
			oauth2.auth.schema.get("flow") === "authorization_code"
		) && !oauth2.auth.code) {
			if (!isValid) {
				oauth2.errCb({
					authId: oauth2.auth.name,
					source: "auth",
					level: "warning",
					message: "Authorization may be unsafe, passed state was changed in server. The passed state wasn't returned from auth server."
				});
			}

			if (qp.code) {
				delete oauth2.state;
				oauth2.auth.code = qp.code;
				oauth2.callback({auth: oauth2.auth, redirectUrl: redirectUrl});
			} else {
				let oauthErrorMsg;
				if (qp.error) {
					oauthErrorMsg = "["+qp.error+"]: " +
						(qp.error_description ? qp.error_description+ ". " : "no accessCode received from the server. ") +
						(qp.error_uri ? "More info: "+qp.error_uri : "");
				}

				oauth2.errCb({
					authId: oauth2.auth.name,
					source: "auth",
					level: "error",
					message: oauthErrorMsg || "[Authorization failed]: no accessCode received from the server."
				});
			}
		} else {
			oauth2.callback({auth: oauth2.auth, token: qp, isValid: isValid, redirectUrl: redirectUrl});
		}
		window.close();
	}

	if (document.readyState !== 'loading') {
		run();
	} else {
		document.addEventListener('DOMContentLoaded', function () {
			run();
		});
	}
</script>
</body>
</html>"""

SWAGGER_DOC_PAGE = """<!DOCTYPE html>
<html lang="en">
<head>
	<meta charset="utf-8" />
	<meta name="viewport" content="width=device-width, initial-scale=1" />
	<link type="text/css" rel="stylesheet" href="{swagger_css_url}">
	<title>{title} API Documentation</title>
	<style>
		code {{ tab-size: 4; }}
		pre {{ tab-size: 4; }}
	</style>
</head>
<body>
<div id="swagger-ui"></div>
<script src="{swagger_js_url}"></script>
<script>
window.onload = () => {{
	window.ui = SwaggerUIBundle({{
		url: '{openapi_url}',
		dom_id: '#swagger-ui',
		presets: [
			SwaggerUIBundle.presets.apis,
			SwaggerUIBundle.SwaggerUIStandalonePreset
		]
	}})
}}
</script>
</body>
</html>"""



import logging
import unittest

from bspump.asab.abc.singleton import Singleton
from .pipeline import UnitTestPipeline
from ..application import BSPumpApplication
from ..abc.processor import Processor


class TestCase(unittest.TestCase):
    def setUp(self) -> None:
        self.App = BSPumpApplication(args=[])

    def tearDown(self):
        Singleton.delete(self.App.__class__)
        self.App = None
        root_logger = logging.getLogger()
        root_logger.handlers = []


class ProcessorTestCase(TestCase):
    """
    A class whose instances are single processor test cases.

    Test authors should subclass ProcessorTestCase for their own tests. Construction
    and deconstruction of the test's environment ('fixture') can be
    implemented by overriding the 'setUp' and 'tearDown' methods respectively.

    See :class: `unittest.TestCase` for more details

    Example of use:

    .. code-block:: python

            class MyProcessorTestCase(ProcessorTestCase)

                    def test_my_processor(self):

                            # setup processor for test
                            self.set_up_processor(my_project.processors.MyProcessor)

                            output = self.execute(
                                    [(None, {'foo': 'bar'})]  # Context, event
                            )

                            self.assertEqual(
                                    [event for context, event in output],
                                    [{'FOO': 'BAR'}]
                            )

    """

    def set_up_processor(self, processor: type(Processor), *args, **kwargs) -> None:
        """
        Construct Pipeline from processor and appends it to PumpService

        :param processor: Processor you want to test
        :param args: Optional arguments for processor
        :param kwargs: Optional key-word arguments for processor
        """

        svc = self.App.get_service("bspump.PumpService")

        self.Pipeline = UnitTestPipeline(self.App, processor, *args, **kwargs)
        svc.add_pipeline(self.Pipeline)

    def execute(self, input_data: []):
        """
        Executes ProcessorTestCase

        You can define custom mocks between calling `set_up_processor` and `execute`

        :return: `input_data` processed by testing processor
        """

        self.Pipeline.Source.Input = input_data

        # TODO catch AttributeError: 'TestIteratorSource' object has no attribute 'Pipeline'
        # Add help text - did you forget to `set_up_processor`?

        self.Pipeline.unittest_start()
        self.App.run()

        return self.Pipeline.Sink.Output



import bspump.common
import bspump.unittest
import bspump.trigger


class TestIteratorGenerator(bspump.unittest.ProcessorTestCase):
    def test_iterator_generator(self):
        event_message = "Generate me!"
        events = {
            (None, event_message),
        }

        self.set_up_processor(bspump.common.IteratorGenerator)

        output = self.execute(events)

        self.assertEqual(
            [event for context, event in output], [i for i in event_message]
        )


class TestIteratorSource(bspump.unittest.ProcessorTestCase):
    def test_iterator_source(self):
        # TODO Generalize to SourceTestCase
        svc = self.App.get_service("bspump.PumpService")

        iterator = iter([1, 2, 3])

        pipeline = bspump.Pipeline(app=self.App)
        sink = bspump.unittest.UnitTestSink(self.App, pipeline)
        source = bspump.common.IteratorSource(self.App, pipeline, iterator).on(
            bspump.trigger.PubSubTrigger(self.App, "Application.run!", self.App.PubSub)
        )
        pipeline.build(source, sink)

        pipeline.PubSub.subscribe("bspump.pipeline.cycle_end!", self._on_finished)

        svc.add_pipeline(pipeline)
        source.TriggerEvent.set()
        self.App.run()

        self.assertEqual([({}, 1), ({}, 2), ({}, 3)], sink.Output)

    def _on_finished(self, event_name, pipeline):
        self.App.stop()



from .attributefilter import AttributeFilter
from .contentfilter import ContentFilter
from .timedriftfilter import TimeDriftFilter

__all__ = [
    "AttributeFilter",
    "ContentFilter",
    "TimeDriftFilter",
]



from ...abc import Expression


class LOOKUP_GET(Expression):
    Attributes = {
        "Key": ["*"],  # TODO: This ...
        "LookupID": ["*"],  # TODO: This ...
    }

    def __init__(self, app, *, arg_in, arg_what):
        super().__init__(app)
        self.PumpService = app.get_service("bspump.PumpService")
        self.LookupID = arg_in
        self.Key = arg_what

    def build_key(self, context, event, *args, **kwargs):
        if isinstance(self.Key, (list, tuple)):
            # Compound key
            result = list()
            for k in self.Key:
                if isinstance(k, (str, int, float)):
                    result.append(k)
                else:
                    result.append(k(context, event, **kwargs))
            return tuple(result)
        else:
            # Simple key
            if isinstance(self.Key, (str, int, float)):
                return self.Key
            else:
                return self.Key(context, event, **kwargs)

    def __call__(self, context, event, *args, **kwargs):
        lookup = self.PumpService.locate_lookup(self.LookupID, context)
        key = self.build_key(context, event, *args, **kwargs)
        return lookup.get(key)


class LOOKUP_CONTAINS(LOOKUP_GET):
    Attributes = {
        "Key": ["*"],  # TODO: This ...
        "LookupID": ["*"],  # TODO: This ...
    }

    def __call__(self, context, event, *args, **kwargs):
        lookup = self.PumpService.locate_lookup(self.LookupID, context)
        key = self.build_key(context, event, *args, **kwargs)
        return lookup.get(key) is not None



import re
import urllib.parse

from ..value.valueexpr import VALUE
from ...abc import Expression


class RegexABCParser(object):
    def __init__(self, app, regex):
        self.Regex = re.compile(regex)

    def __call__(self, context, value):
        return dict(self.Regex.findall(value))


class KeyValueSpaceParser(RegexABCParser):
    def __init__(self, app):
        super().__init__(app, r"\b(\w+)=(.*?)(?=\s\w+=\s*|$)")


class KeyValueDoubleQuotesSpaceParser(RegexABCParser):
    def __init__(self, app):
        super().__init__(app, r'\b(\w+)="(.*?)(?="\s\w+=\s*|"$)')


class QueryStringParser(object):
    def __init__(self, app):
        pass

    def __call__(self, context, value):
        return dict(urllib.parse.parse_qsl(value))


class DICT_PARSE(Expression):
    Attributes = {
        "Value": ["*"],  # TODO: This ...
        "Type": ["*"],  # TODO: This ...
        "Unset": [],  # TODO: This ...
        "Update": ["*"],  # TODO: This ...
    }

    def __init__(
        self,
        app,
        *,
        arg_what,
        arg_type=None,
        arg_set=None,
        arg_unset=None,
        arg_update=None
    ):
        super().__init__(app)

        self.Value = arg_what
        self.Type = arg_type
        self.Parser = {
            "kvs": KeyValueSpaceParser,
            "kvdqs": KeyValueDoubleQuotesSpaceParser,
            "qs": QueryStringParser,
        }.get(arg_type, "kvs")(app)

        self.ArgSet = arg_set
        self.Set = None

        if arg_unset is not None:
            assert isinstance(arg_unset, list)
        self.Unset = arg_unset

        self.Update = arg_update

    def set(self, key, value):
        setattr(self, key, value)

        if "Set" in key:
            self.Set[key[3:]] = value

    def initialize(self):
        if self.ArgSet is None:
            self.Set = None
        else:
            assert isinstance(self.ArgSet, dict)
            self.Set = dict()
            self._set_value_or_expression_to_attribute(self.ArgSet, self.Set, "Set")

    def _set_value_or_expression_to_attribute(self, _from, _to, name):
        for key, value in _from.items():
            # Parse value to attribute, if not expression
            if isinstance(value, Expression):
                _to[key] = value
            else:
                assert (
                    isinstance(value, (int, str, bytes, bool, tuple, list))
                    or value is None
                )
                value = VALUE(self.App, value=value)
                _to[key] = value

            # Set the value to the class attribute
            attr_name = "{}{}".format(name, key)
            setattr(self, attr_name, value)
            self.Attributes[attr_name] = value.get_outlet_type()

    def __call__(self, context, event, *args, **kwargs):
        value = self.Value(context, event, *args, **kwargs)
        ret = self.Parser(context, value)

        if self.Set is not None:
            for key, value in self.Set.items():
                v = value(context, event, ret, *args, **kwargs)
                if v is not None:
                    ret[key] = v

        if self.Update is not None:
            update_dict = self.Update(context, event, ret, *args, **kwargs)
            if update_dict is not None and update_dict is not False:
                ret.update(update_dict)

        if self.Unset is not None:
            for key in self.Unset:
                ret.pop(key, None)

        return ret



#!/usr/bin/env python3
import logging
import io
import fastavro
from ..avro import loader
from ..abc.generator import Generator

L = logging.getLogger(__name__)

###


class AvroSerializer(Generator):
    ConfigDefaults = {
        "schema_file": "",
        "max_block_size": 10,
    }

    def __init__(self, app, pipeline, id=None, config=None):
        super().__init__(app, pipeline, id=id, config=config)
        self.App = app
        self.Schema = loader.load_avro_schema(self.Config)
        self.MaxBlockSize = self.Config["max_block_size"]
        self.Records = []
        self.Context = None
        self.Depth = None
        self.App.PubSub.subscribe("Application.tick!", self.on_tick)

    # TODO: call this method also on_tick() to ensure proactive flushing of accumulated events

    async def on_tick(self, event_name):
        await self.do_generate()

    async def generate(self, context, event, depth):
        self.Context = context
        self.Depth = depth
        self.Event = event
        await self.do_generate()

    async def do_generate(self):
        self.Records.append(self.Event)
        if len(self.Records) < self.MaxBlockSize:
            return None

        records = self.Records
        self.Records = []

        fo = io.BytesIO()

        if self.Schema is None:
            L.warning("Schema file is not provided.")
        else:
            L.warning("Schema file is used.")

        fastavro.writer(fo, self.Schema, records)
        self.Pipeline.inject(self.Context, fo.getbuffer(), self.Depth)



import logging
import aiohttp
import copy

import bspump.asab as asab

#

L = logging.getLogger(__name__)

#


class HTTPTarget(asab.Configurable):
    def __init__(self, svc, config_section_name, config=None):
        super().__init__(config_section_name, config)
        self.URL = self.Config.get("url")

    async def process(self, metrics, now):
        metrics_to_send = copy.deepcopy(metrics)
        async with aiohttp.ClientSession() as session:
            async with session.post(self.URL, json=metrics_to_send) as resp:
                response = await resp.text()
                if resp.status != 200:
                    L.warning(
                        "Error when sending metrics by HTTPTarget: {}\n{}".format(
                            resp.status, response
                        )
                    )



import bspump.analyzer
import bspump.unittest
from unittest.mock import patch


class TestTimeDriftAnalyzer(bspump.unittest.ProcessorTestCase):
    @patch("asab.application.Application.time")
    def test_time_drift_analyzer(self, mocked_time):
        mocked_time.return_value = 100
        events = [
            (None, {"@timestamp": 30}),
            (None, {"foo": "bar"}),
        ]
        self.set_up_processor(bspump.analyzer.TimeDriftAnalyzer)

        output = self.execute(events)

        self.assertEqual(
            [event for context, event in output], [{"@timestamp": 30}, {"foo": "bar"}]
        )
        self.assertEqual(self.Pipeline.Processor.History, [70])



import logging
from bspump.asab import Config

from bspump.asab.zookeeper import build_client
from bspump.abc.lookupprovider import LookupBatchProviderABC

###

L = logging.getLogger(__name__)

###


class ZooKeeperBatchLookupProvider(LookupBatchProviderABC):
    """
    Fetches lookup data from given zookeeper URL.
    """

    def __init__(self, lookup, url, id=None, config=None):
        super().__init__(lookup, url, id, config)
        self.ZKClient, self.Path = build_client(Config, self.URL)

    async def load(self):
        try:
            await self.ZKClient.start()
        except Exception as e:
            L.error("Cannot start zookeeper client: {}".format(e))
            return None
        try:
            data = await self.ZKClient.get_data(self.Path)
        except Exception as e:
            L.error("Failed to fetch '{}': {}".format(self.Path, e))
            data = None
        finally:
            await self.ZKClient.close()
        return data



import os

import aiohttp.web

from .. import Config
from ..web.rest import json_response
from ..web.auth import noauth


class APIWebHandler(object):
    def __init__(self, api_svc, webapp, log_handler):
        self.App = api_svc.App
        self.ApiService = api_svc

        # Add routes
        webapp.router.add_get("/asab/v1/environ", self.environ)
        webapp.router.add_get("/asab/v1/config", self.config)

        webapp.router.add_get("/asab/v1/logs", log_handler.get_logs)
        webapp.router.add_get("/asab/v1/logws", log_handler.ws)

        webapp.router.add_get("/asab/v1/changelog", self.changelog)
        webapp.router.add_get("/asab/v1/manifest", self.manifest)

    @noauth
    async def changelog(self, request):
        """
        It returns a change log file.
        ---
        tags: ['asab.api']
        """

        if self.ApiService.ChangeLog is None:
            return aiohttp.web.HTTPNotFound()

        with open(self.ApiService.ChangeLog, "r") as f:
            result = f.read()

        return aiohttp.web.Response(text=result, content_type="text/markdown")

    @noauth
    async def manifest(self, request):
        """
        It returns the manifest of the ASAB service.

        THe manifest is a JSON object loaded from `MANIFEST.json` file.
        The manifest contains the creation (build) time and the version of the ASAB service.
        The `MANIFEST.json` is produced during the creation of docker image by `asab-manifest.py` script.

        Example of `MANIFEST.json`:

        ```
        {
                'created_at': 2022-03-21T15:49:37.14000,
                'version' :v22.9-4
        }
        ```

        ---
        tags: ['asab.api']
        """

        if self.ApiService.Manifest is None:
            return aiohttp.web.HTTPNotFound()

        return json_response(request, self.ApiService.Manifest)

    @noauth
    async def environ(self, request):
        """
        It returns a JSON response containing the contents of the environment variables.

        Example:

        ```
        {
                "LANG": "en_GB.UTF-8",
                "SHELL": "/bin/zsh",
                "HOME": "/home/foobar",
        }

        ```

        ---
        tags: ['asab.api']
        """
        return json_response(request, dict(os.environ))

    @noauth
    async def config(self, request):
        """
        It returns the JSON with the config of the ASAB service.

        IMPORTANT: All passwords are erased.

        Example:

        ```
        {
                "general": {
                        "config_file": "",
                        "tick_period": "1",
                        "uid": "",
                        "gid": ""
                },
                "asab:metrics": {
                        "native_metrics": "true",
                        "expiration": "60"
                }
        }
        ```

        ---
        tags: ['asab.api']
        """

        # Copy the config and erase all passwords
        result = {}
        for section in Config.sections():
            result[section] = {}
            # Access items in the raw mode (they are not interpolated)
            for option, value in Config.items(section, raw=True):
                if section == "passwords":
                    result[section][option] = "***"
                else:
                    result[section][option] = value
        return json_response(request, result)



import bspump.asab as asab


class ProcessorBase(asab.Configurable):
    """
    Description:

    |

    """

    def __init__(self, app, pipeline, id=None, config=None):
        """
        Initializes the Parameters

        **Parameters**

        app : object
                        Application object.

        pipeline : :meth:`Pipeline <bspump.Pipeline()>`
                        Name of the :meth:`Pipeline <bspump.Pipeline()>`.

        id : str, default=None,
                        ID of the class of config.

        config : JSON, or other compatible formats, default=None
                        Configuration file.


        """
        super().__init__(
            "pipeline:{}:{}".format(
                pipeline.Id, id if id is not None else self.__class__.__name__
            ),
            config=config,
        )

        self.App = app
        self.Loop = app.Loop

        self.Id = id if id is not None else self.__class__.__name__
        self.Pipeline = pipeline

        self.EventCount = 0

    def time(self):
        """
        Accurate representation of a time in the :meth:`Pipeline <bspump.Pipeline()>`.

        :return: App.time()

        """
        return self.App.time()

    @classmethod
    def construct(cls, app, pipeline, definition: dict):
        """
        Can construct a :meth:`processor <bspump.Processor()>` based on a specific definition. For example, a JSON file.

        **Parameters**

        app : Application
                        Name of the `Application <https://asab.readthedocs.io/en/latest/asab/application.html#>_`.

        pipeline : str
                        Name of the :meth:`Pipeline <bspump.Pipeline()>`.

        definition : dict
                        Set of instructions based on which :meth:`processor <bspump.Processor()>` can be constructed.


        :return: cls(app, pipeline, id=newid, config=config)

        |

        """
        newid = definition.get("id")
        config = definition.get("config")
        args = definition.get("args")
        if args is not None:
            return cls(app, pipeline, id=newid, config=config, **args)
        else:
            return cls(app, pipeline, id=newid, config=config)

    def process(self, context, event):
        """
        Can be implemented to return event based on a given logic.

        **Parameters**

        context :
                        Additional information passed to the method.

        event : Data with time stamp stored in any data type, usually it is in JSON.
                        You can specify an event that is passed to the method.

        """
        raise NotImplementedError()

    def locate_address(self):
        """
        Returns an ID of a :meth:`processor <bspump.Processor()>` and a :meth:`Pipeline <bspump.Pipeline()>`.

        :return: ID of the :meth:`Pipeline <bspump.Pipeline()>` and self.Id.

        |

        """
        return "{}.{}".format(self.Pipeline.Id, self.Id)

    def rest_get(self):
        """
        Description:

        :return:
        """
        return {
            "Id": self.Id,
            "Class": self.__class__.__name__,
            "PipelineId": self.Pipeline.Id,
        }

    def __repr__(self):
        return "%s(%r)" % (self.__class__.__name__, self.locate_address())


class Processor(ProcessorBase):
    """
    Inherits from ProcessorBase.

    |

    """

    pass



import logging
import functools
import inspect

import bspump.asab.exceptions as exceptions

#

L = logging.getLogger(__name__)

#


def require(*resources):
    """
    Specify resources required for endpoint access.
    Requests without these resources result in HTTP 403 response.

    Args:
            resources (Iterable): Resources required to access the decorated method.

    Examples:

    ```python
    @asab.web.authz.require("my-app:token:generate")
    async def generate_token(self, request):
            data = await self.service.generate_token()
            return asab.web.rest.json_response(request, data)
    ```
    """

    def decorator_require(handler):
        @functools.wraps(handler)
        async def wrapper(*args, **kwargs):
            request = args[-1]

            if not hasattr(request, "has_resource_access"):
                raise Exception(
                    "Cannot check resource access. Make sure that AuthService is installed and that "
                    "the handler method does not use both the '@noauth' and the '@require' decorators at once."
                )

            if not request.has_resource_access(*resources):
                raise exceptions.AccessDeniedError()

            return await handler(*args, **kwargs)

        return wrapper

    return decorator_require


def noauth(handler):
    """
    Exempt the decorated handler from authentication and authorization.
    The `tenant`, `user_info` and `resources` arguments are not available in the handler.

    Examples:

    ```python
    @asab.web.authz.noauth
    async def get_public_info(self, request):
            data = await self.service.get_public_info()
            return asab.web.rest.json_response(request, data)
    ```
    """
    argspec = inspect.getfullargspec(handler)
    args = set(argspec.kwonlyargs).union(argspec.args)
    for arg in ("tenant", "user_info", "resources"):
        if arg in args:
            raise Exception(
                "{}(): Handler with @noauth cannot have {!r} in its arguments.".format(
                    handler.__qualname__, arg
                )
            )
    handler.NoAuth = True

    @functools.wraps(handler)
    async def wrapper(*args, **kwargs):
        return await handler(*args, **kwargs)

    return wrapper



import logging
import csv
import ipaddress

from bspump.abc.lookup import DictionaryLookup

###

L = logging.getLogger(__name__)

###


class IPGeoLookup(DictionaryLookup):
    """
    This lookup performs transformation of IP address into a geographical location.
    It uses a file database from ip2location.com.
    Lookup provides locations both in ipv4 and ipv6 formats.
    NOTICE: IPv6 database includes also all IPv4 locations, see "ipv4mapped" config. option.

    Free versions: IP2LOCATION-LITE-DB5.IPV6.CSV and IP2LOCATION-LITE-DB5.IPV4.CSV
    For better precision visit https://lite.ip2location.com to buy a commercial version of database.

    Usage: specify in configuration the path to the database in csv format."""

    ConfigDefaults = {
        "path": "",
        "ipv4mapped": "no",  # IPv4-mapped IPv6 address (enables to use IPv6 lookups for IPv4 addresses)
    }

    def __init__(self, app, id=None, config=None):
        super().__init__(app, id=id, config=config)
        self.TreeRoot = None
        self.Locations = {}
        if self.Config["ipv4mapped"].lower() == "yes":
            self.IP4Mapped = True
        else:
            self.IP4Mapped = False

    async def load(self):
        fname = self.Config["path"]
        if fname == "":
            return

        with open(fname, "r") as f:
            array = []

            for line in csv.reader(f, delimiter=","):
                ip_int_address_start = int(line[0])
                ip_int_address_end = int(line[1])
                array.append(ip_int_address_start)
                array.append(ip_int_address_end)
                lat = float(line[6])
                lon = float(line[7])

                if (lat == 0.0) or (lon == 0.0):
                    d = {"lat": None, "lon": None}
                else:
                    d = {"lat": lat, "lon": lon}

                if line[2] != "-":
                    d["country"] = line[2]
                if line[4] != "-":
                    d["region"] = line[4]
                if line[5] != "-":
                    d["city"] = line[5]

                self.Locations[ip_int_address_start] = d
                self.Locations[ip_int_address_end] = d

        self.TreeRoot = self.sorted_array_to_bst(array)
        del array
        L.debug("IPGeoLookup {} was successfully created".format(self.Id))
        return True

    def set(self, tree):
        self.TreeRoot = tree

    # REST

    def rest_get(self):
        rest = super().rest_get()
        rest["TreeRoot"] = self.TreeRoot
        rest["Locations"] = self.Locations
        rest["IP4Mapped"] = self.IP4Mapped
        return rest

    def sorted_array_to_bst(self, arr):
        if not arr:
            return None

        mid = int(len(arr) / 2)
        root = Node(arr[mid])
        root.left = self.sorted_array_to_bst(arr[:mid])
        root.right = self.sorted_array_to_bst(arr[mid + 1 :])
        return root

    def search(self, value):
        root = self.TreeRoot
        while True:
            if root.data == value:
                return root.data

            if root.data > value:
                left = root.left
                if left is None:
                    return root.data
                else:
                    root = left
            if root.data < value:
                right = root.right
                if right is None:
                    return root.data
                else:
                    root = right

    def lookup_location_ipv4(self, address):
        if self.TreeRoot is None:
            # L.warning("Cannot enrich the location")
            return None

        address_int = int(ipaddress.IPv4Address(address))

        if self.IP4Mapped:
            # https://blog.ip2location.com/knowledge-base/ipv4-mapped-ipv6-address/
            # 191.239.213.197 -> ::ffff:191.239.213.197
            address_int += 281470681743360

        value = self.search(address_int)
        return self.Locations.get(value)

    def lookup_location_ipv6(self, address):
        if self.TreeRoot is None:
            # L.warning("Cannot enrich the location")
            return None

        address_int = int(ipaddress.IPv6Address(address))

        value = self.search(address_int)
        return self.Locations.get(value)

    def lookup_location(self, address):
        if "." in address:
            return self.lookup_location_ipv4(address)
        elif ":" in address:
            return self.lookup_location_ipv6(address)
        else:
            raise ValueError("Invalid IPv4/IPv6 format")


class Node:
    def __init__(self, d):
        self.data = d
        self.left = None
        self.right = None



from ..abc.processor import Processor


class FlattenDictProcessor(Processor):
    """
    Description: ....

                    Inspired by https://github.com/amirziai/flatten

            Example:

            "person": {
                    "details": {
                            "first_name": "John",
                            "last_name": "Doe"
                    },
                    "address": {
                            "country": "GB",
                            "city": "London",
                            "postal_code": "WC2N 5DU"
                    }
            }

            Gets converted to:

            {
                    "person.details.first_name": "John",
                    "person.details.last_name": "Doe",
                    "person.address.country": "GB",
                    "person.address.city": "London",
                    "person.address.postal_code": "WC2N 5DU"
            }


    """

    ConfigDefaults = {"separator": "."}

    def __init__(self, app, pipeline, id=None, config=None):
        """
        Description:

        |

        """
        super().__init__(app, pipeline, id, config)
        self.Separator = self.Config["separator"]

    def _construct_key(self, previous_key, new_key):
        """
        Description:

        :return: new_key

        |

        """
        if previous_key:
            return "{}{}{}".format(previous_key, self.Separator, new_key)
        else:
            return new_key

    def flatten(self, nested_dict):
        """
        Description:

        :return: flattened_dict

        |

        """

        flattened_dict = dict()

        def _flatten(object_, key):
            if not object_:
                flattened_dict[key] = object_

            # Dictionary
            elif isinstance(object_, dict):
                for object_key in object_:
                    _flatten(object_[object_key], self._construct_key(key, object_key))
            # Anything else
            else:
                flattened_dict[key] = object_

        _flatten(nested_dict, None)
        return flattened_dict

    def process(self, context, event):
        """
        Description:

        :return: event

        |

        """
        event = self.flatten(event)
        return event



#!/usr/bin/env python3
import json
import fastavro
import logging


###

L = logging.getLogger(__name__)

###


def load_avro_schema(config):
    """
    This module returns a 'schema' if 'schema_file' key is passed in the configuration.

    In the case schema_file is absent it returns 'None'.
    """

    schema_file = config.get("schema_file")

    if schema_file == "":
        return None
    else:
        with open(schema_file, "r") as fi:
            schema = json.load(fi)
        return fastavro.parse_schema(schema)



import logging

from ..analyzer.timewindowanalyzer import TimeWindowAnalyzer


###

L = logging.getLogger(__name__)

###


class TimeSeriesPredictor(TimeWindowAnalyzer):
    """
    Trained model based time window analyzer, which collects
    data into time series and predicts certain value.
    """

    ConfigDefaults = {"path": "", "predicted_attribute": "predicted"}

    def __init__(
        self,
        app,
        pipeline,
        model,
        matrix_id=None,
        dtype=[("value", "f8"), ("predicted", "f8"), ("count", "i8")],
        columns=15,
        analyze_on_clock=False,
        resolution=60,
        start_time=None,
        clock_driven=False,
        id=None,
        config=None,
    ):
        super().__init__(
            app,
            pipeline,
            matrix_id=matrix_id,
            dtype=dtype,
            columns=columns,
            analyze_on_clock=analyze_on_clock,
            resolution=resolution,
            start_time=start_time,
            clock_driven=clock_driven,
            id=id,
            config=config,
        )

        self.Model = model
        self.PredictedAttribute = self.Config["predicted_attribute"]
        self.initialize_window()

    def initialize_window(self):
        """
        Specific initialization if needed.
        """
        pass

    def enrich(self, context, event, predicted):
        """
        Enriches event with predicted value, override if needed.
        """
        event[self.PredictedAttribute] = predicted

    def assign(self, *args):
        """
        Record predicted values into time window matrix (or anywhere) if needed.
        """
        pass

    def process(self, context, event):
        if self.predicate(context, event):
            sample, column = self.evaluate(context, event)
        else:
            return event

        if sample is not None:
            transformed_sample = self.Model.transform(sample)
            predicted = self.Model.predict(transformed_sample)
            self.enrich(context, event, predicted)
            self.assign(predicted, column)
        else:
            self.enrich(context, event, None)

        return event

    def alarm(self, *args):
        """
        Compare real value and predicted and raise an alarm.
        """
        pass



# THIS-FILE-WILL-BE-REPLACED !!! DO NOT CHANGE OR MODIFY THIS FILE !!!

# During the GitHub release workflow, this file is overwritten with the CalVer version
# For development, this defaults to "devel"

# PEP 440 -- Version Identification and Dependency Specification
# https://www.python.org/dev/peps/pep-0440/

__version__ = "2025.10.20.21"
__build__ = "0.0.0.dev"



from unittest.mock import patch

import bspump.unittest
from bspump.kafka import KafkaConnection, KafkaSink


class TestKafkaSink(bspump.unittest.ProcessorTestCase):
    @patch("bspump.kafka.sink.KafkaSink._on_health_check", lambda *x: None)
    def test_producer_params_configuration(self):
        svc = self.App.get_service("bspump.PumpService")
        svc.add_connection(
            KafkaConnection(
                self.App, "KafkaConnection", config={"bootstrap_servers": "kafka:9092"}
            )
        )

        self.set_up_processor(
            KafkaSink,
            connection="KafkaConnection",
            config={
                "client_id": "a",
                "metadata_max_age_ms": "1",
                "request_timeout_ms": "2",
                "api_version": "2.9.0",
                "max_batch_size": "3",
                "max_request_size": "4",
                "linger_ms": "5",
                "send_backoff_ms": "6",
                "retry_backoff_ms": "7",
                "connections_max_idle_ms": "8",
                "enable_idempotence": "True",
                "transactional_id": "c",
                "transaction_timeout_ms": "9",
                "acks": "all",
            },
        )

        output = self.execute([(None, {})])
        self.assertEqual(output, [])

        self.assert_config_value("client_id", "a")
        self.assert_config_value("metadata_max_age_ms", 1)
        self.assert_config_value("request_timeout_ms", 2)
        self.assert_config_value("api_version", "2.9.0")
        self.assert_config_value("max_batch_size", 3)
        self.assert_config_value("max_request_size", 4)
        self.assert_config_value("linger_ms", 5)
        self.assert_config_value("send_backoff_ms", 6)
        self.assert_config_value("retry_backoff_ms", 7)
        self.assert_config_value("connections_max_idle_ms", 8)
        self.assert_config_value("enable_idempotence", True)
        self.assert_config_value("transactional_id", "c")
        self.assert_config_value("transaction_timeout_ms", 9)
        self.assert_config_value("acks", "all")

    def assert_config_value(self, key, value):
        self.assertEqual(value, self.Pipeline.Processor._producer_params.get(key))



import logging
import signal
import sys
import os

from bspump.asab import Application, Config
import bspump.asab.web
import bspump.asab.api

from .service import BSPumpService
from .__version__ import __version__, __build__

L = logging.getLogger(__name__)


class BSPumpApplication(Application):
    """
    Description: BSPumpApplication is **class** used for .....

    :return:
    """

    def __init__(self, args=None):
        super().__init__(args=args)

        # Banner
        print("BitSwan BSPump version {}".format(__version__))

        from bspump.asab.proactor import Module

        self.add_module(Module)

        from bspump.asab.metrics import Module

        self.add_module(Module)

        self.ASABApiService = bspump.asab.api.ApiService(self)

        self.PumpService = BSPumpService(self)
        self.WebContainer = None

        from bspump.asab.alert import AlertService

        self.AlertService = AlertService(self)

        self.MQTTService = None

        try:
            # Signals are not available on Windows
            self.Loop.add_signal_handler(signal.SIGUSR1, self._on_signal_usr1)
        except (NotImplementedError, AttributeError):
            pass

        # Register bspump API endpoints, if requested (the web service is present)
        if "web" in Config and Config["web"].get("listen"):
            # Initialize API service
            self.add_module(bspump.asab.web.Module)

            self.WebService = self.get_service("asab.WebService")

            from .web import initialize_web

            self.WebContainer = initialize_web(self.WebService.WebContainer)
            self.ASABApiService.initialize_web()

        mqtt_username = os.environ.get("MQTT_USERNAME")
        mqtt_password = os.environ.get("MQTT_PASSWORD")
        mqtt_broker_url = os.environ.get("MQTT_BROKER_URL")

        if mqtt_broker_url and self.DeploymentId and (mqtt_username or mqtt_password):
            from .mqtt import MQTTService, MQTTConnection

            self.PumpService.add_connection(
                MQTTConnection(
                    self,
                    "MQTTServiceConnection",
                    {
                        "username": mqtt_username,
                        "password": mqtt_password,
                        "broker": mqtt_broker_url,
                    },
                )
            )
            self.MQTTService = MQTTService(self, connection="MQTTServiceConnection")

        # Initialize zookeeper container
        if "zookeeper" in Config.sections():
            from bspump.asab.zookeeper import Module

            self.add_module(Module)

            self.ASABApiService.initialize_zookeeper()

    def create_argument_parser(self):
        """
        Description:

        :return:
        """
        prog = sys.argv[0]
        if prog[-11:] == "__main__.py":
            prog = sys.executable + " -m bspump"

        description = """
BSPump is a stream processor. It is a part of BitSwan.
For more information, visit: https://github.com/LibertyAces/BitSwanPump

version: {}
build: {} [{}]
""".format(
            __version__, __build__, __build__[:7]
        )
        parser = super().create_argument_parser(prog=prog, description=description)
        parser.add_argument(
            "--test", action="store_true", help="Run pipeline/automation tests"
        )
        parser.add_argument(
            "notebook", nargs="?", default=None, help="Jupyter notebook"
        )
        # add watch argument that watches the notebooks for changes and restarts if needed
        parser.add_argument(
            "--watch",
            action="store_true",
            help="Watch the notebook for changes and restarts if needed",
        )
        return parser

    def parse_arguments(self, args=None):
        args = super().parse_arguments(args=args)
        self.Test = args.test
        self.Notebook = args.notebook
        if self.Notebook is None:
            self.Notebook = os.environ.get("JUPYTER_NOTEBOOK", "pipelines/main.ipynb")
        self.Watch = args.watch

    async def main(self):
        print("{} pipeline(s) ready.".format(len(self.PumpService.Pipelines)))

        if self.Watch:
            from .watch import Watcher

            Watcher([self.Notebook])

    def _on_signal_usr1(self):
        """
        Description:

        :return:

        :hint: To clear reset from all pipelines, run
        $ kill -SIGUSR1 xxxx
        Equivalently, you can use `docker kill -s SIGUSR1 ....` to reset containerized BSPump.
        """
        # Reset errors from all pipelines
        for pipeline in self.PumpService.Pipelines.values():
            if not pipeline.is_error():
                continue  # Focus only on pipelines that has errors
            pipeline.set_error(None, None, None)



import logging
import os
import json
from asab.timer import Timer
import pyarrow as pa
import pyarrow.parquet as pq
import pandas as pd
from bspump.abc.sink import Sink

#

L = logging.getLogger(__file__)

#


class ParquetSink(Sink):
    """
    schema_file is JSON which defines column names (as keys) and their types.
    These types are allowed : string, bool, float, int, list, decimal
                                                                            date, time, bytearray, array

    schema_file example:
            {
                    "name":{
                            "type": "string"
                    },
                    "age":{
                            "type": "int"
                    },
                    "salary":{
                            "type": "float"
                    }
            }
    """

    ConfigDefaults = {
        "rows_in_chunk": 1000,
        "rollover_mechanism": "rows",  # or time
        "rows_per_file": 10000,
        "writing_period": "1d",  # only used if rollover_mechanism == 'time', valid units are s, m, h, d, w
        "file_name_template": "./sink{index}.parquet",
    }

    def __init__(self, app, pipeline, id=None, config=None):
        super().__init__(app, pipeline, id=id, config=config)
        app.PubSub.subscribe("Application.tick/10!", self.on_tick)

        metrics_service = app.get_service("asab.MetricsService")
        self.Counter = metrics_service.create_counter(
            "counter", tags={}, init_values={"parquet.error": 0}
        )
        self.Gauge = metrics_service.create_gauge(
            "gauge",
            tags={},
            init_values={"parquet.missing_attributes": 0, "parquet.new_attributes": 0},
        )

        self.Frames = []
        self.ChunkSize = self.Config["rows_in_chunk"]
        self.Index = 0
        self.RolloverMechanism = self.Config["rollover_mechanism"]
        self.FileNameTemplate = self.Config["file_name_template"]

        self.MissingSet = set()
        self.NewSet = set()

        self.SchemaFile = self.Config.get("schema_file")
        if self.SchemaFile is not None:
            self.SchemaDefined = True
        else:
            self.SchemaDefined = False

        self.SchemaDataTypes = [
            "string",
            "bool",
            "list",
            "decimal",
            "float",
            "int",
            "date",
            "time",
            "bytearray",
            "array",
        ]

        if self.SchemaDefined is True:
            with open(self.SchemaFile) as json_data:
                schema = json.load(json_data)
                schema_is_valid = self.schema_validator(schema)
                print(str(schema_is_valid))
                if schema_is_valid:
                    self.Schema = schema
                else:
                    self.SchemaDefined = False

        if self.RolloverMechanism == "rows":
            if self.Config["rows_per_file"] % self.ChunkSize != 0:
                self.ChunksPerFile = int(
                    round(self.Config["rows_per_file"] / self.ChunkSize)
                )
                L.warning(
                    "rows_per_file % rows_in_chunk needs to be zero. rows_per_file was rounded"
                )
            elif self.Config["rows_per_file"] < self.ChunkSize:
                self.ChunksPerFile = 1
                L.warning(
                    "rows_per_file needs to be equal or greater than rows_in_chunk, rows_per_file changed to rows_in_chunk value"
                )
            else:
                self.ChunksPerFile = int(self.Config["rows_per_file"] / self.ChunkSize)
            self.Chunks = 0

        elif self.RolloverMechanism == "time":
            writing_period = self.Config["writing_period"]

            seconds_per_unit = {"s": 1, "m": 60, "h": 3600, "d": 86400, "w": 604800}

            def convert_to_seconds(s):
                return int(s[:-1]) * seconds_per_unit[s[-1]]

            writing_period = convert_to_seconds(writing_period)

            self._writing_timer = Timer(app, self.rotate_async, autorestart=True)
            self._writing_timer.start(writing_period)

        else:
            self.Index = None
            self.Chunks = None

        self._pq_writer = None

    def schema_validator(self, schema):
        if schema is not None:
            for col in schema.values():
                if col["type"] is None:
                    L.warning(
                        "SchemaFile is not valid schema. Type not found for every column."
                    )
                    return False

                col["type"] = col["type"].lower()

                if col["type"] not in self.SchemaDataTypes:
                    L.warning(
                        "SchemaFile is not valid schema. Types must be one of the following: string, bool, \
								float, int, list, decimal, date, time, bytearray or array. Not {}".format(
                            col["type"]
                        )
                    )

                    return False

            return True

        else:
            L.warning("No SchemaFile specified")
            return False

    def apply_schema(self, event):
        data = {}

        for attr_name, attr_descr in self.Schema.items():
            value = event.pop(attr_name, None)

            if value is None:
                data[attr_name] = None
                self.MissingSet.add(attr_name)

            else:
                value_type = attr_descr["type"]
                try:
                    if value_type == "string" and type(value) is not str:
                        data[attr_name] = str(value)
                        continue

                    if value_type == "int" and type(value) is not int:
                        data[attr_name] = int(float(value))
                        continue

                    if value_type == "float" and type(value) is not float:
                        data[attr_name] = float(value)
                        continue

                    if value_type == "bool" and type(value) is not bool:
                        if type(value) is str and value.lower() == "false":
                            data[attr_name] = False
                            continue
                        else:
                            data[attr_name] = bool(value)
                            continue

                except ValueError:
                    self.Counter.add("parquet.error", 1)
                    pass

                data[attr_name] = value

        new_attrs_count = len(event)
        if new_attrs_count > 0:
            self.NewSet = self.NewSet.union([k for k in event])

        df = pd.DataFrame([data])
        return df

    def build_filename(self, postfix=""):
        return (
            self.FileNameTemplate.format(
                index=("{:04d}".format(self.Index)) if self.Index is not None else ""
            )
            + postfix
        )

    def on_tick(self, event_name):
        self.Gauge.set("parquet.new_attributes", len(self.NewSet))
        self.Gauge.set("parquet.missing_attributes", len(self.MissingSet))
        self.flush()

    def process(self, context, event):
        if self.SchemaDefined:
            df = self.apply_schema(event)
        else:
            df = pd.DataFrame([event])
        self.Frames.append(df)

        if self.RolloverMechanism == "rows" and (len(self.Frames) >= self.ChunkSize):
            self.Chunks = self.Chunks + 1
            if self.Chunks >= self.ChunksPerFile:
                self.rotate()

            self.flush()

    def flush(self):
        if len(self.Frames) != 0:
            table = pa.Table.from_pandas(pd.concat(self.Frames))
            if self._pq_writer is None:
                self._pq_writer = pq.ParquetWriter(
                    self.build_filename("-open"), table.schema
                )
            try:
                self._pq_writer.write_table(table=table)
            except Exception as e:
                L.warning(e)
            self.Frames = []

    async def rotate_async(self):
        self.rotate()

    def rotate(self, new_filename=None):
        """
        Call this to close the currently open file.
        """

        self.flush()
        del self._pq_writer
        self._pq_writer = None

        current_fname = self.build_filename("-open")
        os.rename(current_fname, current_fname[:-5])

        if self.RolloverMechanism == "rows":
            self.Chunks = 0

        self.Index = self.Index + 1



import logging

from ..abc.source import TriggerSource

#

L = logging.getLogger(__name__)

#


class MatrixSource(TriggerSource):
    def __init__(self, app, pipeline, matrix_id, id=None, config=None):
        super().__init__(app, pipeline, id=id, config=config)
        svc = app.get_service("bspump.PumpService")
        self.AnalyzeMatrix = svc.locate_matrix(matrix_id)

    async def cycle(self):
        await self.process(self.AnalyzeMatrix)



from unittest.mock import MagicMock

import bspump.unittest
import bspump.common


class TestStringToBytesParser(bspump.unittest.ProcessorTestCase):
    def test_string_to_bytes_parser(self):
        events = {
            (None, "some string"),
            (None, "another string"),
            (None, "last string"),
        }

        self.set_up_processor(bspump.common.StringToBytesParser)

        output = self.execute(events)

        self.assertEqual(
            sorted([event for context, event in output]),
            [b"another string", b"last string", b"some string"],
        )

    def test_event_not_string(self):
        events = [
            (None, b"Not a string"),
        ]
        self.set_up_processor(bspump.common.StringToBytesParser)

        output = self.execute(events)

        self.assertTrue(self.Pipeline.is_error())
        self.assertEqual([event for context, event in output], [])


class TestBytesToStringParser(bspump.unittest.ProcessorTestCase):
    def test_bytes_to_string_parser(self):
        events = {
            (None, b"some bytes"),
            (None, b"another bytes"),
            (None, b"last bytes"),
        }

        self.set_up_processor(bspump.common.BytesToStringParser)

        output = self.execute(events)

        self.assertListEqual(
            sorted([event for context, event in output]),
            ["another bytes", "last bytes", "some bytes"],
        )

    def test_event_not_bytes(self):
        events = [
            (None, "Not a bytes"),
        ]
        self.set_up_processor(bspump.common.BytesToStringParser)

        bspump.pipeline.L = MagicMock()  # turn off logging

        output = self.execute(events)

        self.assertTrue(self.Pipeline.is_error())
        self.assertEqual([event for context, event in output], [])



import os

import bspump.declarative
import bspump.unittest


class TestDeclarativeDictParser(bspump.unittest.TestCase):
    def setUp(self) -> None:
        super().setUp()
        self.Builder = bspump.declarative.ExpressionBuilder(self.App)

    def load(self, decl_fname):
        basedir = os.path.dirname(__file__)
        with open(os.path.join(basedir, decl_fname), "r") as f:
            return self.Builder.parse(f.read())[0]

    def test_dict_parse_kvs_01(self):
        decl = self.load("./test_dict_parse_kvs_01.yaml")
        res = decl(
            {},
            """key1=value1 key2=value2 request=http://www.pandora.com/json/music/artist/justin-moore?explicit\\=false uri=3DLoiRDsBABCAA9FvE1htRg\\=\\=""",
        )
        self.assertEqual(
            res,
            {
                "key1": "value1",
                "key2": "value2",
                "request": "http://www.pandora.com/json/music/artist/justin-moore?explicit\\=false",
                "uri": "3DLoiRDsBABCAA9FvE1htRg\\=\\=",
            },
        )

    def test_dict_parse_kvdqs_01(self):
        decl = self.load("./test_dict_parse_kvdqs_01.yaml")
        res = decl(
            {},
            '''key1="value1" key2="value2" key3="ssss\"aaa" request="http://www.pandora.com/json/music/artist/justin-moore?explicit\\=false"''',
        )
        self.assertEqual(
            res,
            {
                "key1": "value1",
                "key2": "value2",
                "key3": 'ssss"aaa',
                "request": "http://www.pandora.com/json/music/artist/justin-moore?explicit\\=false",
            },
        )

    def test_dict_parse_qs_01(self):
        decl = self.load("./test_dict_parse_qs_01.yaml")
        res = decl({}, """first=this+is+a+field&second=was+it+clear+%28already%29%3F""")
        self.assertEqual(
            res, {"first": "this is a field", "second": "was it clear (already)?"}
        )



import bspump.common
from bspump import Processor, Pipeline
from bspump.trigger import PubSubTrigger
from bspump.unittest import UnitTestSource, UnitTestSink


class RaisingProcessor(Processor):
    def process(self, context, event):
        if "warning" in event or "error" in event:
            raise Exception()
        return event


class WarningPipeline(Pipeline):
    def __init__(self, app, id=None, config=None):
        super().__init__(app, id, config)
        self.PubSub.subscribe("bspump.pipeline.cycle_end!", self._on_finished)
        self.Source = UnitTestSource(app, self).on(
            PubSubTrigger(app, "Application.run!", app.PubSub)
        )
        self.Processor = RaisingProcessor(app, self)
        self.Sink = UnitTestSink(app, self)
        self.build(self.Source, self.Processor, self.Sink)

    def handle_error(self, exception, context, event):
        if event == "warning":
            return True
        return not super().handle_error(exception, context, event)

    def _on_finished(self, event_name, pipeline):
        self.App.stop()


class TestMetricsService(bspump.unittest.ProcessorTestCase):
    # TODO: "Application.Metrics.Flush!" does not exist in asab anymore

    def _metric_flush(self, _, __, counter):
        if counter.get("error"):
            self.ErrorCount += counter["error"]
        if counter.get("warning"):
            self.WarningCount += counter["warning"]

    def test_metrics_service(self):
        self.ErrorCount = 0
        self.WarningCount = 0

        self.App.PubSub.subscribe("Application.Metrics.Flush!", self._metric_flush)

        svc = self.App.get_service("bspump.PumpService")

        self.Pipeline = WarningPipeline(self.App)

        self.Pipeline.Source.Input = [
            (None, "ok"),
            (None, "warning"),
            (None, "error"),
        ]
        self.Pipeline.PubSub.publish("unittest.go!")
        svc.add_pipeline(self.Pipeline)
        self.App.run()

        self.assertEqual([({}, "ok")], self.Pipeline.Sink.Output)
        self.assertEqual(0, self.ErrorCount)



import asyncio
import logging

import aiopg
import aiopg.utils

import psycopg2
import psycopg2.errorcodes
import psycopg2.extras

from bspump.asab import PubSub
from ..abc.connection import Connection


L = logging.getLogger(__name__)


class PostgreSQLConnection(Connection):
    ConfigDefaults = {
        "host": "127.0.0.1",
        "port": 5432,
        "user": "",
        "password": "",
        "db": "",
        "connect_timeout": 60,
        "reconnect_delay": 5.0,
        "output_queue_max_size": 10,
        "max_bulk_size": 1,  # This is because execute many is not supported by aiopg
    }

    # This is just guess work based on the existing MySQL retryables
    # This doesn't work for connection errors for sync_connection
    RetryErrors = frozenset(
        [
            psycopg2.errorcodes.ADMIN_SHUTDOWN,
            psycopg2.errorcodes.CRASH_SHUTDOWN,
            psycopg2.errorcodes.CANNOT_CONNECT_NOW,
            psycopg2.errorcodes.CONNECTION_FAILURE,
            psycopg2.errorcodes.CONNECTION_EXCEPTION,
            psycopg2.errorcodes.SQLSERVER_REJECTED_ESTABLISHMENT_OF_SQLCONNECTION,
            psycopg2.errorcodes.SQLCLIENT_UNABLE_TO_ESTABLISH_SQLCONNECTION,
            psycopg2.errorcodes.CLASS_INVALID_TRANSACTION_INITIATION,
            psycopg2.errorcodes.TOO_MANY_CONNECTIONS,
            psycopg2.errorcodes.QUERY_CANCELED,
        ]
    )

    def __init__(self, app, id=None, config=None):
        super().__init__(app, id=id, config=config)

        self.ConnectionEvent = asyncio.Event()
        self.ConnectionEvent.clear()

        self.PubSub = PubSub(app)
        self.Loop = app.Loop

        self._host = self.Config["host"]
        self._port = int(self.Config["port"])
        self._user = self.Config["user"]
        self._password = self.Config["password"]
        self._connect_timeout = self.Config["connect_timeout"]
        self._db = self.Config["db"]
        self._reconnect_delay = self.Config["reconnect_delay"]
        self._output_queue_max_size = self.Config["output_queue_max_size"]
        self._max_bulk_size = int(self.Config["max_bulk_size"])

        self._conn_sync = None
        self._conn_future = None
        self._connection_request = False
        self._pause = False

        # Subscription
        self._on_health_check("connection.open!")
        app.PubSub.subscribe("Application.stop!", self._on_application_stop)
        app.PubSub.subscribe("Application.tick!", self._on_health_check)
        app.PubSub.subscribe("PostgreSQLConnection.pause!", self._on_pause)
        app.PubSub.subscribe("PostgreSQLConnection.unpause!", self._on_unpause)

        self._output_queue = asyncio.Queue()
        self._bulks = {}  # We have a "bulk" per query

    def _on_pause(self):
        self._pause = True

    def _on_unpause(self):
        self._pause = False

    def _flush(self):
        for query in self._bulks.keys():
            # Break if throttling was requested during the flush,
            # so that put_nowait doesn't raise
            if self._pause:
                break

            self._flush_bulk(query)

    def _flush_bulk(self, query):
        # Enqueue and thorttle if needed
        self._output_queue.put_nowait((query, self._bulks[query]))
        if self._output_queue.qsize() == self._output_queue_max_size:
            self.PubSub.publish("PostgreSQLConnection.pause!", self)

        # Reset bulk
        self._bulks[query] = []

    def _on_application_stop(self, message_type, counter):
        self._flush()
        self._output_queue.put_nowait((None, None))

    def _on_health_check(self, message_type):
        if self._conn_future is not None:
            # Connection future exists

            if not self._conn_future.done():
                # Connection future didn't result yet
                # No sanitization needed
                return

            try:
                self._conn_future.result()
            except Exception:
                # Connection future threw an error
                L.exception("Unexpected connection future error")

            # Connection future already resulted (with or without exception)
            self._conn_future = None

        assert self._conn_future is None

        self._conn_future = asyncio.ensure_future(
            self._async_connection(), loop=self.Loop
        )

        # Connection future already resulted (with or without exception)
        self._sync_connection()

    def build_dsn(self):
        dsn = ""
        # Database
        # if len(self._db) > 0:
        dsn += "dbname={} ".format(self._db)

        # Host
        # if len(self._host) > 0:
        dsn += "host={} ".format(self._host)

        # Port
        # if self._port is not None:
        dsn += "port={} ".format(self._port)

        # User
        if len(self._user) > 0:
            dsn += "user={} ".format(self._user)

        # Password
        if len(self._password):
            dsn += "password={} ".format(self._password)

        dsn = dsn.strip()
        return dsn

    async def _async_connection(self):
        dsn = self.build_dsn()

        try:
            async with aiopg.create_pool(
                dsn=dsn, timeout=self._connect_timeout, loop=self.Loop
            ) as pool:
                self._conn_pool = pool
                self.ConnectionEvent.set()

                await self._loader()

        except (
            psycopg2.OperationalError,
            psycopg2.ProgrammingError,
            psycopg2.InternalError,
        ) as e:
            if e.pgcode in self.RetryErrors:
                L.warning(
                    "Recoverable error '{}' ({}) occurred in PostgreSQLConnection.".format(
                        e.pgerror, e.pgcode
                    )
                )
                return None
            raise e
        except BaseException as e:
            L.exception("Unexpected PostgresSQL connection error. %s" % e)
            raise

    def acquire(self) -> aiopg.pool._PoolConnectionContextManager:
        """
                Acquire asynchronous database connection

                Use with `async with` statement

        .. code-block:: python

                async with self.Connection.acquire_connection() as connection:
                        async with connection.cursor() as cursor:
                                await cursor.execute(query)

                :return: Asynchronous Context Manager
        """
        assert self._conn_pool is not None
        return self._conn_pool.acquire()

    def _sync_connection(self):
        """
        Acquire synchronous psycopg2 connection
        Currently only used for iterating over a PostgreSQLLookup
        """
        dsn = self.build_dsn()
        if self._conn_sync is not None:
            try:
                self._conn_sync.close()
            except (
                psycopg2.OperationalError,
                psycopg2.ProgrammingError,
                psycopg2.InternalError,
            ):
                pass
            self._conn_sync = None
        try:
            connection = psycopg2.connect(
                dsn=dsn, connect_timeout=self._connect_timeout
            )
            self._conn_sync = connection
        except (
            psycopg2.OperationalError,
            psycopg2.ProgrammingError,
            psycopg2.InternalError,
        ) as e:
            if e.pgcode in self.RetryErrors:
                L.warning(
                    "Recoverable error '{}' ({}) occurred in PostgreSQLConnection.".format(
                        e.pgerror, e.pgcode
                    )
                )
                return None
            L.exception("Unexpected PostgreSQL (psycopg2) connection error")
            raise e
        except BaseException:
            L.exception("Unexpected PostgreSQL (psycopg2) connection error")

            raise

    def acquire_sync_cursor(self) -> psycopg2.extras.DictCursor:
        """
        Acquire synchronous psycopg2 cursor
        Currently only used for iterating over a PostgreSQLLookup
        """
        assert self._conn_sync is not None
        return self._conn_sync.cursor(cursor_factory=psycopg2.extras.DictCursor)

    def consume(self, query, data):
        # Create a bulk for this query if doesn't yet exist
        if query not in self._bulks:
            self._bulks[query] = []

        # Add data to the query's bulk
        self._bulks[query].append(data)

        # Flush on _max_bulk_size
        if len(self._bulks[query]) >= self._max_bulk_size:
            self._flush_bulk(query)

    async def _loader(self):
        while True:
            query, data = await self._output_queue.get()

            if query is None:
                break

            if self._output_queue.qsize() == self._output_queue_max_size - 1:
                self.PubSub.publish(
                    "PostgreSQLConnection.unpause!", self, asynchronously=True
                )

            try:
                async with self.acquire() as conn:
                    try:
                        async with conn.cursor() as cur:
                            for item in data:
                                _query = cur.mogrify(query, item)
                                await cur.execute(_query)
                    except BaseException:
                        L.exception(
                            "Unexpected error when processing PostgreSQL query."
                        )
            except BaseException:
                L.exception("Couldn't acquire connection")



import logging
import ldap
import ldap.controls

from ..abc.source import TriggerSource

#

L = logging.getLogger(__name__)

#


class LDAPSource(TriggerSource):
    """
    Runs an LDAP query and pushes the resulting events down the pipeline.

    All output values are bytestrings!
    """

    ConfigDefaults = {
        "base": "dc=example,dc=org",
        "filter": "(&(objectClass=inetOrgPerson)(cn=*))",
        "attributes": "dn objectGUID sAMAccountName email givenName sn UserAccountControl",
        "results_per_page": 1000,
    }

    def __init__(self, app, pipeline, connection, id=None, config=None):
        super().__init__(app, pipeline, id=id, config=config)
        self.Connection = pipeline.locate_connection(app, connection)
        self.ProactorService = app.get_service("asab.ProactorService")
        self.Scope = ldap.SCOPE_SUBTREE
        self.Base = self.Config.get("base")
        self.Filter = self.Config.get("filter")
        self.Attributes = self.Config.get("attributes").split(" ")
        self.ResultsPerPage = self.Config.getint("results_per_page")

    async def cycle(self):
        # TODO: Throttling
        await self.Pipeline.ready()
        cookie = b""
        while True:
            page, cookie = await self.ProactorService.execute(
                self._search_worker, cookie
            )
            for entry in page:
                await self.process(entry, context={})
            if cookie is None or len(cookie) == 0:
                break

    def _search_worker(self, cookie=b""):
        page = []
        with self.Connection.ldap_client() as client:
            paged_results_control = ldap.controls.SimplePagedResultsControl(
                True, size=self.ResultsPerPage, cookie=cookie
            )
            msgid = client.search_ext(
                base=self.Base,
                scope=self.Scope,
                filterstr=self.Filter,
                attrlist=self.Attributes,
                serverctrls=[paged_results_control],
            )
            res_type, res_data, res_msgid, serverctrls = client.result3(msgid)
            for dn, attrs in res_data:
                if dn is None:
                    # Skip system entries
                    continue

                event = {}
                # LDAP returns all attributes as lists of bytestrings, e.g.:
                #   {"sAMAccountName": [b"vhavel"], ...}
                # Unpack them
                for k, v in attrs.items():
                    if isinstance(v, list):
                        if len(v) < 1:
                            continue
                        elif len(v) == 1:
                            v = v[0]
                    event[k] = v
                page.append(event)

            for sc in serverctrls:
                if (
                    sc.controlType
                    == ldap.controls.SimplePagedResultsControl.controlType
                ):
                    cookie = sc.cookie
                    break
            else:
                L.error("No SimplePagedResultsControl cookie in result serverctrls")
                cookie = b""

            return page, cookie



import json
import logging

import requests

from ..abc.lookup import MappingLookup
from ..abc.lookup import AsyncLookupMixin
from ..cache import CacheDict


L = logging.getLogger(__name__)


class ElasticSearchLookup(MappingLookup, AsyncLookupMixin):
    """
    The lookup that is linked with a ES.
    It provides a mapping (dictionary-like) interface to pipelines.
    It feeds lookup data from ES using a query.
    It also has a simple cache to reduce a number of database hits.

    **configs**

    *index* - Elastic's index

    *key* - field name to match

    *timefield* - field name to use for sorting

    *sort_order* - order of sorting (default is 'desc')

    *scroll_timeout* - Timeout of single scroll request (default is '1m'). Allowed time units:
    https://www.elastic.co/guide/en/elasticsearch/reference/current/common-options.html#time-units

    Example:

    .. code:: python

                    The ElasticSearchLookup can be then located and used inside a custom enricher:

                                    class AsyncEnricher(bspump.Generator):

                                                    def __init__(self, app, pipeline, id=None, config=None):
                                                                    super().__init__(app, pipeline, id, config)
                                                                    svc = app.get_service("bspump.PumpService")
                                                                    self.Lookup = svc.locate_lookup("MySQLLookup")

                                                    async def generate(self, context, event, depth):
                                                                    if 'user' not in event:
                                                                                    return None

                                                                    info = await self.Lookup.get(event['user'])

                                                                    # Inject a new event into a next depth of the pipeline
                                                                    self.Pipeline.inject(context, event, depth)

    """

    ConfigDefaults = {
        "index": "",  # Specify an index
        "key": "",  # Specify field name to match
        "cache_non_existent": False,  # True - if key is not found, it will be cached as None
        "sort_order": "desc",
        "scroll_timeout": "1m",
    }

    def __init__(self, app, connection, id=None, config=None, cache=None, lazy=False):
        """
        Description:

        **Parameters**

        app : Application
                        Name of the Application.

        connection : Connection
                        Name of the Connection

        id : ID, default= None
                        ID

        config : JSON, default= None
                        Configuration file with additional information.

        cache : ?,default= None

        lazy : ?, default= None

        """
        super().__init__(app, id=id, config=config, lazy=lazy)
        self.Connection = connection

        self.Index = self.Config["index"]
        self.ScrollTimeout = self.Config["scroll_timeout"]
        self.Key = self.Config["key"]
        self.CacheNonExistent = {"True": True, "False": False}.get(
            self.Config["cache_non_existent"], False
        )
        self.Timefield = self.Config.get("timefield")
        self.SortOrder = self.Config["sort_order"]

        self.Count = -1
        if cache is None:
            self.Cache = CacheDict()
        else:
            self.Cache = cache

        metrics_service = app.get_service("asab.MetricsService")
        self.CacheCounter = metrics_service.create_counter(
            "es.lookup.cache", tags={}, init_values={"hit": 0, "miss": 0}
        )
        self.SuccessCounter = metrics_service.create_counter(
            "es.lookup.success", tags={}, init_values={"hit": 0, "miss": 0}
        )

    async def _find_one(self, key):
        prefix = "_search"
        request = {"size": 1, "query": self.build_find_one_query(key)}
        if self.Timefield:
            request["sort"] = [{self.Timefield: self.SortOrder}]

        url = self.Connection.get_url() + "{}/{}".format(self.Index, prefix)

        async with self.Connection.get_session() as session:
            async with session.post(
                url, json=request, headers={"Content-Type": "application/json"}
            ) as response:
                if response.status != 200:
                    data = await response.text()
                    L.error(
                        "Failed to fetch data from ElasticSearch: {} from {}\n{}".format(
                            response.status, url, data
                        )
                    )

                msg = await response.json()
                try:
                    hit = msg["hits"]["hits"][0]
                except Exception:
                    return None

        return hit["_source"]

    async def get(self, key):
        """
        Obtain the value from lookup asynchronously.

        **Parameters**

        key : ?


        :return: value

        |

        """
        value = None
        try:
            value = self.Cache[key]
            self.CacheCounter.add("hit", 1)
        except KeyError:
            try:
                value = await self._find_one(key)
                if value is not None:
                    self.Cache[key] = value
                    self.CacheCounter.add("miss", 1)
                elif self.CacheNonExistent:
                    self.Cache[key] = None
                    self.CacheCounter.add("miss", 1)
            except Exception as e:
                L.warn("There was an exception {}".format(e))

        if value is None:
            self.SuccessCounter.add("miss", 1)
        else:
            self.SuccessCounter.add("hit", 1)

        return value

    def build_find_one_query(self, key) -> dict:
        """
        Override this method to build your own lookup query

        **Parameters**

        key : ?

        :return: Default single-key query

        |

        """
        return {"match": {self.Key: key}}

    async def _count(self):
        prefix = "_count"
        request = {"query": {"match_all": {}}}

        url = self.Connection.get_url() + "{}/{}".format(self.Index, prefix)

        async with self.Connection.get_session() as session:
            async with session.post(
                url, json=request, headers={"Content-Type": "application/json"}
            ) as response:
                if response.status != 200:
                    data = await response.text()
                    L.error(
                        "Failed to fetch data from ElasticSearch: {} from {}\n{}".format(
                            response.status, url, data
                        )
                    )

                msg = await response.json()

        return int(msg["count"])

    async def load(self):
        """
        Sets the length of Cache to Count.

        :return: True

        |

        """
        self.Count = len(self.Cache)
        return True

    def __len__(self):
        return self.Count

    def __getitem__(self, key):
        # To avoid synchronous operations completely
        raise NotImplementedError()

    def __iter__(self):
        scroll_id = None
        request = {"size": 10000, "query": {"match_all": {}}}

        all_hits = []
        while True:
            if scroll_id is None:
                path = "{}/_search?scroll={}".format(self.Index, self.ScrollTimeout)
                request_body = request
            else:
                path = "_search/scroll"
                request_body = {"scroll": self.ScrollTimeout, "scroll_id": scroll_id}

            url = self.Connection.get_url() + path
            response = requests.post(url, json=request_body)

            if response.status_code != 200:
                data = response.text()
                L.error(
                    "Failed to fetch data from ElasticSearch: {} from {}\n{}".format(
                        response.status_code, url, data
                    )
                )
                break

            data = json.loads(response.text)

            scroll_id = data.get("_scroll_id")

            if scroll_id is None:
                break

            hits = data["hits"]["hits"]

            if len(hits) == 0:
                break

            all_hits.extend(hits)

        self.Iterator = all_hits.__iter__()
        return self

    def __next__(self):
        element = next(self.Iterator)

        key = element["_source"].get(self.Key)
        if key is not None:
            self.Cache[key] = element["_source"]
        return key

    @classmethod
    def construct(cls, app, definition: dict):
        """
        Constructs config, id, and connection based on config.

        **Parameters**

        app : Application
                        Name of the Application.

        definition:dict : Definition
                        Definition containing information about certain variables.

        :return: cls(app, newid, connection, config)

        |

        """
        newid = definition.get("id")
        config = definition.get("config")
        connection = definition["args"]["connection"]
        return cls(app, newid, connection, config)



import os
import os.path
import sys
import re
import glob
import logging
import inspect
import platform
import configparser
import urllib.parse
import collections.abc
import typing
from dotenv import load_dotenv

from . import utils


L = logging.getLogger(__name__)


class ConfigParser(configparser.ConfigParser):
    """
    ConfigParser enhanced with new features such as adding default configuration, URL validation, automatic reading from Zookeeper etc.
    """

    _syslog_sockets = {"Darwin": "/var/run/syslog"}

    _syslog_format = {"Darwin": "m"}

    _default_values = {
        "general": {
            "config_file": os.environ.get("ASAB_CONFIG", ""),
            "tick_period": 1,  # In seconds
            "var_dir": os.path.expanduser(
                "~/." + os.path.splitext(os.path.basename(sys.argv[0]))[0]
            ),
            "changelog": "",
            "manifest": "",
            # Daemonization
            "pidfile": "!",  # '!' has a special meaning => it transforms into platform specific location of pid file
            "working_dir": ".",
            "uid": "",
            "gid": "",
        },
        "asab:metrics": {
            "native_metrics": "true",
            "web_requests_metrics": False,  # False is a default, web_requests_metrics won't be generated.
            "expiration": 60,
        },
        "asab:doc": {"default_route_tag": "module_name"},
        "logging": {
            "verbose": os.environ.get("ASAB_VERBOSE", False),
            "app_name": os.path.basename(sys.argv[0]),
            "sd_id": "sd",  # Structured data id, see RFC5424
            "level": "NOTICE",
            "levels": "",
        },
        "logging:console": {
            "format": "%(asctime)s %(levelname)s %(name)s %(struct_data)s%(message)s",
            "datefmt": "%d-%b-%Y %H:%M:%S.%f",
        },
        "logging:syslog": {
            "enabled": "false",
            # TODO: "facility": 'local1',
            "address": _syslog_sockets.get(platform.system(), "/dev/log"),
            "format": _syslog_format.get(platform.system(), "3"),
        },
        "logging:file": {
            "path": "",
            "format": "%(asctime)s %(levelname)s %(name)s %(struct_data)s%(message)s",
            "datefmt": "%d-%b-%Y %H:%M:%S.%f",
            "backup_count": 3,
            "backup_max_bytes": 0,
            "rotate_every": "",
        },
        "library": {
            "azure_cache": "false",  # true or the actual path of where the cache should be located
        },
        # "passwords" section serves to securely store passwords
        # in the configuration file; the passwords are not
        # shown in the default API
        #
        # Usage in the configuration file:
        #
        # [connection:KafkaConnection]
        # password=${passwords:kafka_password}
        #
        # [passwords]
        # kafka_password=<MY_SECRET_PASSWORD>
        "passwords": {},
        "housekeeping": {
            "at": "03:00",
            "limit": "05:00",
            "run_at_startup": "no",
        },
    }

    if "ASAB_ZOOKEEPER_SERVERS" in os.environ:
        # If `ASAB_ZOOKEEPER_SERVERS` are specified, use that as a default value
        _default_values["zookeeper"] = {"servers": os.environ["ASAB_ZOOKEEPER_SERVERS"]}

    def add_defaults(self, dictionary: dict) -> None:
        """Add defaults to a current configuration.

        Args:
                dictionary: Arguments to be added to the default configuration.
        """
        # sniff file for secrets groups so we can read the env vars
        config_file = self._default_values.get("general", {}).get("config_file", "")
        if not config_file:
            config_file = "pipelines.conf"
        if config_file and os.path.exists(config_file):
            temp_config = ConfigParser()
            temp_config.read(config_file)
            if groups := temp_config.get("secrets", "groups", fallback=""):
                gitops_dir = os.environ.get(
                    "BITSWAN_GITOPS_DIR",
                    os.path.join(
                        os.environ.get("HOME"), ".config/bitswan/local-gitops/"
                    ),
                )
                if type(groups) is not list:
                    groups = groups.split(" ")
                for group in groups:
                    if group:
                        load_dotenv(os.path.join(gitops_dir, "secrets", group))

        for section, keys in dictionary.items():
            section = str(section)

            if section not in self._sections:
                try:
                    self.add_section(section)
                except ValueError:
                    if self._strict:
                        raise

            for key, value in keys.items():
                key = self.optionxform(str(key))
                if key in self._sections[section]:
                    # Value exists, no default needed
                    continue

                if value is not None:
                    value = str(value)

                if value is not None and "$" in value:
                    self.set(section, key, os.path.expandvars(value))
                else:
                    self.set(section, key, value)

    def _traverse_includes(self, includes: str, this_dir: str) -> None:
        """
        Read included config files. Nested including is supported.
        """
        if "\n" in includes:
            sep = "\n"
        else:
            sep = " "

        for include_glob in includes.split(sep):
            include_glob = include_glob.strip()

            if len(include_glob) == 0:
                continue

            if include_glob.startswith("zookeeper"):
                self._include_from_zookeeper(include_glob)

            include_glob = os.path.expandvars(include_glob.strip())

            for include in glob.glob(include_glob):
                include = os.path.abspath(include)

                if include in self._included:
                    # Preventing infinite dependency looping
                    L.warn(
                        "Config file '{}' can be included only once.".format(include)
                    )
                    continue

                self._included.add(include)
                self.set("general", "include", "")

                self._load_dir_stack.append(os.path.dirname(include))
                try:
                    self.read(include)
                finally:
                    self._load_dir_stack.pop()

                includes = self.get("general", "include", fallback="")
                self._traverse_includes(includes, os.path.dirname(include_glob))

    def _load(self):
        """
        This method should be called only once, any subsequent call will lead to undefined behaviour.
        """
        self._load_dir_stack = []
        self.config_contents_list = []
        self.config_name_list = []

        config_fname = ConfigParser._default_values["general"]["config_file"]

        if config_fname == "":
            config_fname = "pipelines.conf"
            if not os.path.isfile(config_fname):
                config_fname = ""
        else:
            if not os.path.isfile(config_fname):
                print(
                    "Config file '{}' not found".format(config_fname), file=sys.stderr
                )
                sys.exit(1)

        if config_fname:
            self._load_dir_stack.append(os.path.dirname(config_fname))
            try:
                self.read(config_fname)
            finally:
                self._load_dir_stack.pop()

        self.add_defaults(ConfigParser._default_values)

        includes = self.get("general", "include", fallback="")

        self._included = set()
        self._traverse_includes(includes, this_dir=os.path.dirname(config_fname))

        del self._load_dir_stack

    def _include_from_zookeeper(self, zkurl):
        """
        Load the configuration from a ZooKeeper server and append it to the `self.config_contents_list` attribute.

        The method establishes a connection to the ZooKeeper server specified in the configuration file mentioned above.
        It retrieves the configuration by accessing the path specified in the `general` section, using the key `includes`.
        The server URL is provided as a list of server names: server1, server2, server3.
        The path to the configuration file follows this format: 'zookeeper-1:2181,zookeeper-2:2181,zookeeper-3:2181/asab/etc/zk-site.conf.'

        The loaded configuration is then appended to the `self.config_contents_list` attribute, allowing further processing or usage.
        This method supports loading configuration files in various formats, such as .json, .yaml, and .conf.

        Example:

                ```ini
                [asab:zookeeper]
                url=server1 server2 server3
                [general]
                include=zookeeper://zookeeper-1:2181,zookeeper-2:2181,zookeeper-3:2181/asab/etc/zk-site.conf.
                ```
        """
        # parse include value into hostname and path
        url_pieces = urllib.parse.urlparse(zkurl)
        url_path = url_pieces.path
        url_netloc = url_pieces.netloc

        if not url_netloc:
            if "asab:zookeeper" in self:
                # Backward compatibility
                url_netloc = self["asab:zookeeper"]["servers"]
            else:
                url_netloc = self["zookeeper"]["servers"]

        if url_path.startswith("./"):
            if "asab:zookeeper" in self:
                # Backward compatibility
                url_path = self["asab:zookeeper"]["path"] + url_path[1:]
            else:
                url_path = self["zookeeper"]["path"] + url_path[1:]

        head, tail = os.path.split(url_path)
        self.config_name_list.append(tail)

        try:
            # Delayed import to minimize a hard dependency footprint
            import kazoo.client
            import json
            import yaml

            zk = kazoo.client.KazooClient(url_netloc)
            zk.start()
            data = zk.get(url_path)[0]
            if url_path.endswith(".json"):
                config = json.loads(data)
                self.read_dict(config)
            elif url_path.endswith(".yaml"):
                config = yaml.safe_load(data)
                self.read_dict(config)
            elif url_path.endswith(".conf"):
                config = data.decode("utf-8")
                self.read_string(config)
            else:
                raise NotImplementedError(
                    "Unknown configuration format '{}'".format(url_path)
                )

            zk.stop()
            zk.close()

            # Include in the list of config file contents
            self.config_contents_list.append(config)

        except Exception as e:
            print(
                "Failed to obtain configuration from Zookeeper server(s): '{}'.".format(
                    e
                )
            )
            sys.exit(1)

    def get_config_contents_list(self):
        return self.config_contents_list, self.config_name_list

    def getseconds(
        self, section, option, *, raw=False, vars=None, fallback=None, **kwargs
    ) -> float:
        """
        Get time data from config and convert time string into seconds with `convert_to_seconds()` method.

        The available units are:

        - `y` - years
        - `M` - months
        - `w` - weeks
        - `d` - days
        - `h` - hours
        - `m` - minutes
        - `s` - seconds
        - `ms` - milliseconds

        Returns:
                float: Time in seconds.

        Examples:

        ```python
        self.SleepTime = asab.Config["sleep"].getseconds("sleep_time")
        self.AnotherSleepTime = asab.Config.getseconds("sleep", "another_sleep_time")
        ```
        """

        if fallback is None:
            fallback = configparser._UNSET

        return self._get_conv(
            section,
            option,
            utils.convert_to_seconds,
            raw=raw,
            vars=vars,
            fallback=fallback,
            **kwargs,
        )

    def geturl(
        self,
        section,
        option,
        *,
        raw=False,
        vars=None,
        fallback=None,
        scheme=None,
        **kwargs,
    ):
        """
        Get URL from config and remove all leading and trailing whitespaces and trailing slashes.

        Args:
                scheme (str | tuple): URL scheme(s) awaited. If `None`, scheme validation is bypassed.

        Returns:
                Validated URL.

        Raises:
                ValueError: Scheme requirements are not met if set.

        Examples:

        ```ini
        [urls]
        teskalabs=https://www.teskalabs.com/
        github=github.com
        ```

        ``` python
        asab.Config["urls"].geturl("teskalabs", scheme="https")
        asab.Config.geturl("urls", "github", scheme=None)
        ```
        """
        return utils.validate_url(
            self.get(section, option, raw=raw, vars=vars, fallback=fallback), scheme
        )

    def getmultiline(
        self, section, option, *, raw=False, vars=None, fallback=None, **kwargs
    ) -> typing.List[str]:
        """
        Get multiline data from config.

        Examples:

        ```ini
        [places]
        visited:
                Praha
                Brno
                Pardubice Plzeň
        unvisited:
        ```

        ```python
        >>> asab.Config.getmultiline("places", "visited")
        ["Praha", "Brno", "Pardubice", "Plzeň"]
        >>> asab.Config.getmultiline("places", "unvisited")
        []
        >>> asab.Config.getmultiline("places", "nonexisting", fallback=["Gottwaldov"])
        ["Gottwaldov"]
        ```
        """
        values = self.get(section, option, raw=raw, vars=vars, fallback=fallback)
        if isinstance(values, str):
            return [item.strip() for item in re.split(r"\s+", values) if len(item) > 0]
        else:
            # fallback can be anything
            return values


class _Interpolation(configparser.ExtendedInterpolation):
    """Interpolation which expands environment variables in values."""

    def before_read(self, parser, section, option, value):
        # Expand environment variables
        if "$" in value:
            os.environ["THIS_DIR"] = os.path.abspath(parser._load_dir_stack[-1])

            value = os.path.expandvars(value)

        return super().before_read(parser, section, option, value)


Config = ConfigParser(interpolation=_Interpolation())
"""
Object for accessing the configuration of the ASAB application.

Examples:

```python
my_conf_value = asab.Config['section_name']['key']
```
"""


class Configurable(object):
    """
    Custom object whose attributes can be loaded from the configuration.

    Example:
            ```python
            class ConfigurableObject(asab.Configurable):

                    ConfigDefaults = {
                            'foo': 'bar',
                    }

                    def __init__(self, config_section_name, config=None):
                            super().__init__(config_section_name=config_section_name, config=config)

                            config_foo = self.Config.get('foo')
            ```
    """

    ConfigDefaults: dict = {}

    def __init__(self, config_section_name: str, config: typing.Optional[dict] = None):
        self.Config = ConfigurableDict()

        for base_class in inspect.getmro(self.__class__):
            if not hasattr(base_class, "ConfigDefaults"):
                continue
            if len(base_class.ConfigDefaults) == 0:
                continue

            # Merge config defaults of each base class in the 'inheritance' way
            for key, value in base_class.ConfigDefaults.items():
                if value is None:
                    raise ValueError(
                        "None value not allowed in ConfigDefaults. Found in %s:%s "
                        % (config_section_name, key)
                    )

                if key not in self.Config:
                    self.Config[key] = value

        if Config.has_section(config_section_name):
            for key, value in Config.items(config_section_name):
                self.Config[key] = value

        if config is not None:
            self.Config.update(config)


# This is for backward compatibility
ConfigObject = Configurable


class ConfigurableDict(collections.abc.MutableMapping):
    """
    A dictionary supplemented with custom methods for obtaining bools, seconds, urls etc.
    """

    def __init__(self):
        self._data = {}

    def __getitem__(self, key):
        return self._data[key]

    def __setitem__(self, key, value):
        self._data[key] = value

    def __delitem__(self, key):
        del self._data[key]

    def __iter__(self):
        return iter(self._data)

    def __len__(self):
        return len(self._data)

    def getboolean(self, key) -> bool:
        """
        Obtain the corresponding value of the key and convert it into bool.
        """
        value = self._data[key]
        return utils.string_to_boolean(value)

    def getseconds(self, key) -> float:
        """
        Obtain the corresponding value of the key and convert it into seconds via `convert_to_seconds()` method.
        """
        value = self._data[key]
        return utils.convert_to_seconds(value)

    def getint(self, key) -> int:
        """
        Obtain the corresponding value of the key and convert it into integer.
        """
        value = self._data[key]
        return int(value)

    def getfloat(self, key) -> float:
        """
        Obtain the corresponding value of the key and convert it into float.
        """
        value = self._data[key]
        return float(value)

    def geturl(self, key, scheme):
        """
        Obtain the corresponding value of the key and parse it via `validate_url()` method.
        """
        value = self._data[key]
        return utils.validate_url(value, scheme)

    def __repr__(self):
        return "<%s %r>" % (self.__class__.__name__, self._data)



import abc
from bspump.asab import Configurable


class Connection(abc.ABC, Configurable):
    """
    Connection class is responsible for creating a connection between items or services within the infrastructure of BSPump.
    Their main use is to create connection with the main components of BSPump: source, :meth:`processor <bspump.Processor()>` and sink.

    |

    """

    def __init__(self, app, id=None, config=None):
        """
        Description:


        **Parameters**

        app : Application
                        Specification of an `Application <https://asab.readthedocs.io/en/latest/asab/application.html#>`_.

        id : default None

        config : JSON or other compatible format, default None
                        It contains important information and data responsible for creating a connection.

        """

        _id = id if id is not None else self.__class__.__name__
        super().__init__("connection:{}".format(_id), config=config)

        self.App = app
        self.Loop = app.Loop

        self.Id = _id

    def time(self):
        """
        Returns accurate time of the asynchronous process.

        :hint: More information in the ASAB documentation in `UTC Time <https://asab.readthedocs.io/en/latest/asab/application.html#utc-time>`_.

        |

        """
        return self.App.time()

    @classmethod
    def construct(cls, app, definition: dict):
        """
        Creates a connection based on a specific definition. For example, a JSON file.

        **Parameters**

        app : Application
                        ID of the `Application <https://asab.readthedocs.io/en/latest/asab/application.html#>_`.

        definition : definition format
                        Defines instructions for the method that can be used to create a connection.


        :return: cls(app, newid, config)

        |

        """
        newid = definition.get("id")
        config = definition.get("config")
        return cls(app, newid, config)



import logging
import asyncio
import copy
from ..abc.source import Source
from ..abc.sink import Sink
from ..abc.processor import Processor


L = logging.getLogger(__name__)


class DirectSource(Source):
    """
    Description: This source processes inserted event synchronously.

    |

    """

    def __init__(self, app, pipeline, id=None, config=None):
        """
        Description:

        |

        """
        super().__init__(app, pipeline, id=id, config=config)

    def put(self, context, event, copy_context=False, copy_event=False):
        """
        This method serves to put events into the pipeline and process them right away.

        Context can be an empty dictionary if is not provided.

        |

        """

        if copy_context:
            child_context = {"ancestor": copy.deepcopy(context)}
        else:
            child_context = {"ancestor": context}

        if copy_event:
            event = copy.deepcopy(event)

        # DirectSource is not using the common asynchronous process method
        self.Pipeline.MetricsEPSCounter.add("eps.in", 1)
        self.Pipeline.MetricsCounter.add("event.in", 1)
        self.Pipeline.inject(context=child_context, event=event, depth=0)

    async def main(self):
        """
        Description:

        |

        """
        pass


class InternalSource(Source):
    """
    Description:

    |

    """

    ConfigDefaults = {
        "queue_max_size": 10,  # 0 means unlimited size
        "backpressure": 0.8,  # Percentage of the queue that will result in a backpressure
    }

    def __init__(self, app, pipeline, id=None, config=None):
        """
        Description:

        |

        """
        super().__init__(app, pipeline, id=id, config=config)
        self.Loop = app.Loop

        self.BackPressure = False
        maxsize = int(self.Config.get("queue_max_size"))
        if maxsize == 0:
            self.BackPressureLimit = None
        else:
            if maxsize == 1:
                maxsize = 2  # Special case
            self.BackPressureLimit = maxsize * float(self.Config.get("backpressure"))
            if self.BackPressureLimit == maxsize:
                self.BackPressureLimit -= 1
            assert self.BackPressureLimit > 0
        self.Queue = asyncio.Queue(maxsize=maxsize)

    def put(self, context, event, copy_context=False, copy_event=False):
        """
        Description: Context can be an empty dictionary if is not provided.

        If you are getting a `asyncio.queues.QueueFull` exception,
        you likely did not implemented backpressure handling.
        The simpliest approach is to use RouterSink / RouterProcessor.

        |

        """

        if copy_context:
            context = copy.deepcopy(context)

        if copy_event:
            event = copy.deepcopy(event)

        self.Queue.put_nowait((context, event))

        if (
            (not self.BackPressure)
            and (self.BackPressureLimit is not None)
            and (self.BackPressureLimit <= self.Queue.qsize())
        ):
            self.BackPressure = True
            self.Pipeline.PubSub.publish(
                "bspump.InternalSource.backpressure_on!", source=self
            )

    async def put_async(self, context, event, copy_context=False, copy_event=False):
        """
        Description: This method allows to put an event into InternalSource asynchronously.
        Since a processing in the pipeline is synchronous, this method is useful mainly
        for situation, when an event is created outside of the pipeline processing.
        It is designed to handle situation when the queue is becoming full.

        Context can be an empty dictionary if is not provided.

        |

        """

        if copy_context:
            context = copy.deepcopy(context)

        if copy_event:
            event = copy.deepcopy(event)

        await self.Queue.put((context, event))

        if (
            (not self.BackPressure)
            and (self.BackPressureLimit is not None)
            and (self.BackPressureLimit <= self.Queue.qsize())
        ):
            self.BackPressure = True
            self.Pipeline.PubSub.publish(
                "bspump.InternalSource.backpressure_on!", source=self
            )

    async def main(self):
        """
        Description:

        |

        """
        try:
            while True:
                await self.Pipeline.ready()
                context, event = await self.Queue.get()

                if (
                    (self.BackPressure)
                    and (self.BackPressureLimit is not None)
                    and (self.BackPressureLimit > self.Queue.qsize())
                ):
                    self.BackPressure = False
                    self.Pipeline.PubSub.publish(
                        "bspump.InternalSource.backpressure_off!", source=self
                    )

                await self.process(event, context={"ancestor": context})

                self.Queue.task_done()

        except asyncio.CancelledError:
            if self.Queue.qsize() > 0:
                L.warning(
                    "'{}' stopped with {} events in a queue".format(
                        self.locate_address(), self.Queue.qsize()
                    )
                )

    def rest_get(self):
        """
        Description:

        :return: rest

        |

        """
        rest = super().rest_get()
        rest["Queue"] = self.Queue.qsize()
        rest["BackPressure"] = self.BackPressure
        return rest


class RouterMixIn(object):
    """
    Description: Router Mix in a class

    |

    """

    """
	Router Mix in a class
	"""

    def _mixin_init(self, app):
        """
        Description:

        |

        """
        self.ServiceBSPump = app.get_service("bspump.PumpService")
        self.SourcesCache = {}

    def locate(self, source_id):
        """
        Description:

        :return: source

        |

        """
        source = self.ServiceBSPump.locate(source_id)
        if source is None:
            L.warning("Cannot locate '{}' in '{}'".format(source_id, self.Id))
            raise RuntimeError("Cannot locate '{}' in '{}'".format(source_id, self.Id))

        self.SourcesCache[source_id] = source

        source.Pipeline.PubSub.subscribe(
            "bspump.pipeline.not_ready!", self._on_target_pipeline_ready_change
        )
        source.Pipeline.PubSub.subscribe(
            "bspump.pipeline.ready!", self._on_target_pipeline_ready_change
        )

        # If target pipeline is not ready, throttle
        if not source.Pipeline.is_ready():
            self.Pipeline.throttle(source.Pipeline, enable=True)

        source.Pipeline.PubSub.subscribe(
            "bspump.InternalSource.backpressure_on!",
            self._on_internal_source_backpressure_ready_change,
        )
        source.Pipeline.PubSub.subscribe(
            "bspump.InternalSource.backpressure_off!",
            self._on_internal_source_backpressure_ready_change,
        )
        if source.BackPressure:
            self.Pipeline.throttle(source, enable=True)

        return source

    def unlocate(self, source_id):
        """
        Description: Undo locate() call, it means that it removes the source from a cache + remove throttling binds

        :return: ??

        |

        """

        # UNTESTED CODE !!!
        try:
            source = self.SourcesCache.pop(source_id)
        except KeyError:
            return

        source.Pipeline.PubSub.unsubscribe(
            "bspump.pipeline.not_ready!", self._on_target_pipeline_ready_change
        )
        source.Pipeline.PubSub.unsubscribe(
            "bspump.pipeline.ready!", self._on_target_pipeline_ready_change
        )

        if not source.Pipeline.is_ready():
            self.Pipeline.throttle(source.Pipeline, enable=False)

        source.Pipeline.PubSub.unsubscribe(
            "bspump.InternalSource.backpressure_on!",
            self._on_internal_source_backpressure_ready_change,
        )
        source.Pipeline.PubSub.unsubscribe(
            "bspump.InternalSource.backpressure_off!",
            self._on_internal_source_backpressure_ready_change,
        )

        if source.BackPressure:
            self.Pipeline.throttle(source, enable=False)

    def dispatch(self, context, event, source_id, copy_event=True):
        """
        Description:

        :return: self.route(context, event, source_id, copy_event=True)

        |

        """
        # TODO: Obsolete function
        return self.route(context, event, source_id, copy_event=True)

    def route(self, context, event, source_id, copy_event=True):
        """
        Description: This method routes an event to a InternalSource `source_id`.

        It can be called multiple times from a process() method, which results in a cloning of the event.

        |

        """
        source = self.SourcesCache.get(source_id)

        if source is None:
            source = self.locate(source_id)

        source.put(context, event, copy_event=copy_event)

    def _on_target_pipeline_ready_change(self, event_name, pipeline):
        """
        Description:

        |

        """
        if event_name == "bspump.pipeline.ready!":
            self.Pipeline.throttle(pipeline, enable=False)
        elif event_name == "bspump.pipeline.not_ready!":
            self.Pipeline.throttle(pipeline, enable=True)
        else:
            L.warning(
                "Unknown event '{}' received in _on_target_pipeline_ready_change in '{}'".format(
                    event_name, self
                )
            )

    def _on_internal_source_backpressure_ready_change(self, event_name, source):
        """
        Description:

        |

        """
        if event_name == "bspump.InternalSource.backpressure_off!":
            self.Pipeline.throttle(source, enable=False)
        elif event_name == "bspump.InternalSource.backpressure_on!":
            self.Pipeline.throttle(source, enable=True)
        else:
            L.warning(
                "Unknown event '{}' received in _on_internal_source_backpressure_ready_change in '{}'".format(
                    event_name, self
                )
            )


class RouterSink(Sink, RouterMixIn):
    """
    Description: Abstract Sink that dispatches events to other internal sources.
    One should override the process() method and call route() with target source id.

    |

    """

    def __init__(self, app, pipeline, id=None, config=None):
        super().__init__(app, pipeline, id, config)
        self._mixin_init(app)


class RouterProcessor(Processor, RouterMixIn):
    """
    Description:	Abstract Processor that dispatches events to other internal sources.
    One should override the process() method and call route() with target source id.

    |

    """

    def __init__(self, app, pipeline, id=None, config=None):
        """
        Description:

        |

        """
        super().__init__(app, pipeline, id, config)
        self._mixin_init(app)



import asyncio
import logging
import re
import aiohttp

from ..abc.connection import Connection

#

L = logging.getLogger(__name__)


#


class InfluxDBConnection(Connection):
    """
            Description: InfluxDBConnection serves to connect BSPump application with an InfluxDB database.
            The InfluxDB server is accessed via URL, and the database is specified using the `db` parameter
            in the configuration. Current version of InfluxDB defaultly works with timestamp in epoch nanoseconds.
            If you want to use different epoch format, you must specify it using precision in configuration

    .. code:: python

            app = bspump.BSPumpApplication()
            svc = app.get_service("bspump.PumpService")
            svc.add_connection(
                    bspump.influxdb.InfluxDBConnection(app, "InfluxConnection1")
            )

            **Config Default**

            url : http://localhost:8086/

            db : name of db

            precision : ns (epoch timestamp precision. Default is ns, but can have us,ms,s)

            output_queue_max_size : 10

            output_bucket_max_size : 1000 * 1000

            timout : 30

            retry_enabled : False

            response_codes_to_retry : 404, 502, 503, 504

    """

    ConfigDefaults = {
        "url": "http://localhost:8086/",
        "db": "mydb",
        "precision": "ns",
        "output_queue_max_size": 10,
        "output_bucket_max_size": 1000 * 1000,
        "timeout": 30,
        "retry_enabled": False,
        "response_codes_to_retry": "404,502,503,504",
    }

    def __init__(self, app, id=None, config=None):
        """
        Description:

        **Parameters**

        app : Application
                Name of the Application.

        id : ID, default = None

        config : JSON, default = None
                Configuration file with additional information.

        """
        super().__init__(app, id=id, config=config)

        self.url = self.Config["url"].strip()
        if self.url[-1] != "/":
            self.url += "/"

        self._url_write = (
            self.url
            + "write?db="
            + self.Config["db"]
            + "&precision="
            + self.Config["precision"]
        )

        self._output_bucket_max_size = int(self.Config["output_bucket_max_size"])
        self._output_queue_max_size = int(self.Config["output_queue_max_size"])
        self._timeout = aiohttp.ClientTimeout(total=int(self.Config["timeout"]))
        self.RetryEnabled = self.Config.getboolean("retry_enabled")

        self.AllowedBulkResponseCodes = frozenset(
            [
                int(x)
                for x in re.findall(r"[0-9]+", self.Config["response_codes_to_retry"])
            ]
        )

        self._output_queue = asyncio.Queue()
        self._started = True

        self._output_bucket = ""

        self.PubSub = app.PubSub
        self.PubSub.subscribe("Application.tick!", self._on_tick)
        self.PubSub.subscribe("Application.exit!", self._on_exit)

        self._future = asyncio.ensure_future(self._loader())

    def consume(self, data):
        """
        Description: Consumes user-defined data to be stored in the InfluxDB database.

        **Parameters**

        data :

        """
        self._output_bucket += data
        if len(self._output_bucket) > self._output_bucket_max_size:
            self.flush()

    async def _on_exit(self, event_name):
        self._started = False
        self.flush()
        await self._output_queue.put(
            None
        )  # By sending None via queue, we signalize end of life
        await self._future  # Wait till the _loader() terminates

    async def _on_tick(self, event_name):
        if self._started and self._future.done():
            # Ups, _loader() task crashed during runtime, we need to restart it
            try:
                r = self._future.result()
                # This error should never happen
                L.error(
                    "Influx error observed, returned: '{}' (should be None)".format(r)
                )
            except Exception as e:
                L.exception(f"Influx error, {e}, observed, restoring the order")

            self._future = asyncio.ensure_future(self._loader())
        self.flush()

    def flush(self, event_name=None):
        """
        Description: Directly flushes the content of the internal bucket with data to InfluxDB database.

        **Parameters**

        event_name : ?, default = None

        """
        if len(self._output_bucket) == 0:
            return

        assert self._output_bucket is not None
        self._output_queue.put_nowait(self._output_bucket)
        self._output_bucket = ""

        if self._output_queue.qsize() == self._output_queue_max_size:
            self.PubSub.publish("InfluxDBConnection.pause!", self)

    async def _loader(self):
        # A cycle that regularly sends buckets if there are any
        while self._started:
            _output_bucket = await self._output_queue.get()
            if _output_bucket is None:
                break

            if self._output_queue.qsize() == self._output_queue_max_size - 1:
                self.PubSub.publish(
                    "InfluxDBConnection.unpause!", self, asynchronously=True
                )

            # Sending the data asynchronously

            try:
                async with aiohttp.ClientSession(timeout=self._timeout) as session:
                    async with session.post(
                        self._url_write, data=_output_bucket
                    ) as resp:
                        resp_body = await resp.text()
                        if (
                            resp.status in self.AllowedBulkResponseCodes
                            and self.RetryEnabled
                        ):
                            L.warning(
                                f"Retryable response code recieved, retrying. Queue size {self._output_queue.qsize()}"
                            )
                            self._output_queue.put_nowait(_output_bucket)
                            self.PubSub.publish("InfluxDBConnection.pause!", self)
                            await asyncio.sleep(3)
                            self.PubSub.publish("InfluxDBConnection.unpause!", self)

                        elif resp.status is None:
                            L.error(
                                "Failed to insert a line into Influx status:{} body:{}".format(
                                    resp.status, resp_body
                                )
                            )
                            raise RuntimeError("Failed to insert line into Influx")

            # Here we define errors, that we want to retry
            except OSError:
                if self.RetryEnabled:
                    L.warning(
                        f"Retryable exception raised, retrying. Queue size {self._output_queue.qsize()}"
                    )
                    self._output_queue.put_nowait(_output_bucket)
                    self.PubSub.publish("InfluxDBConnection.pause!", self)
                    await asyncio.sleep(3)
                    self.PubSub.publish("InfluxDBConnection.unpause!", self)



import bspump.analyzer
import bspump.unittest


class TestAnalyzer(bspump.unittest.ProcessorTestCase):
    def test_analyzer(self):
        events = [
            (None, {"attribute": "value"}),
        ]
        self.set_up_processor(bspump.analyzer.Analyzer)

        output = self.execute(events)

        self.assertEqual([event for context, event in output], [{"attribute": "value"}])

    def test_analyzer_on_clock(self):
        events = [
            (None, {"attribute": "value"}),
        ]
        self.set_up_processor(bspump.analyzer.Analyzer, analyze_on_clock=True)

        output = self.execute(events)

        self.assertEqual([event for context, event in output], [{"attribute": "value"}])



import re
import logging
import inspect

from bspump.asab import Config
import aiohttp
import aiohttp.web
import yaml

import typing

from .doc_templates import SWAGGER_OAUTH_PAGE, SWAGGER_DOC_PAGE
from ..web.auth import noauth


##

L = logging.getLogger(__name__)

##


class DocWebHandler(object):
    def __init__(self, api_service, app, web_container, config_section_name="asab:doc"):
        self.App = app
        self.WebContainer = web_container
        self.WebContainer.WebApp.router.add_get("/doc", self.doc)
        self.WebContainer.WebApp.router.add_get(
            "/oauth2-redirect.html", self.oauth2_redirect
        )
        self.WebContainer.WebApp.router.add_get("/asab/v1/openapi", self.openapi)

        self.AuthorizationUrl = Config.get(
            config_section_name, "authorization_url", fallback=None
        )
        self.TokenUrl = Config.get(config_section_name, "token_url", fallback=None)
        self.Scopes = Config.get(config_section_name, "scopes", fallback=None)

        self.Manifest = api_service.Manifest

        self.DefaultRouteTag: str = Config["asab:doc"].get(
            "default_route_tag"
        )  # default: 'module_name'
        if self.DefaultRouteTag not in ["module_name", "class_name"]:
            raise ValueError(
                "Unknown default_route_tag: {}. Choose between options "
                "'module_name' and 'class_name'.".format(self.DefaultRouteTag)
            )

        self.ServerUrls: str = (
            Config.get(config_section_name, "server_urls", fallback="/")
            .strip()
            .split("\n")
        )

    def build_swagger_documentation(self) -> dict:
        """
        Take a docstring of the class and a docstring of methods and merge them into Swagger data.
        """
        app_doc_string: str = self.App.__doc__
        app_description: str = get_docstring_description(app_doc_string)
        specification: dict = {
            "openapi": "3.0.1",
            "info": {
                "title": "{}".format(self.App.__class__.__name__),
                "description": app_description,
                "contact": {
                    "name": "ASAB-based microservice",
                    "url": "https://www.github.com/teskalabs/asab",
                },
                "version": self.get_version_from_manifest(),
            },
            "servers": [{"url": url} for url in self.ServerUrls],
            "components": {"securitySchemes": self.create_security_schemes()},
            # Base path relative to openapi endpoint
            "paths": {},
            # Authorization
            # TODO: Authorization must not be always of OAuth type
        }

        # Application specification
        app_info: dict = self.get_docstring_yaml_dict(app_doc_string)
        if app_info is not None:
            specification.update(app_info)

        # Find asab and microservice routes, sort them alphabetically by the first tag
        asab_routes = []
        microservice_routes = []

        for route in self.WebContainer.WebApp.router.routes():
            if route.method == "HEAD":
                # Skip HEAD methods
                # TODO: once/if there is graphql, its method name is probably `*`
                continue

            # Determine which routes are asab-based
            path: str = self.get_route_path(route)
            if (
                re.search("asab", path)
                or re.search("/doc", path)
                or re.search("/oauth2-redirect.html", path)
            ):
                asab_routes.append(self.parse_route_data(route))
            else:
                microservice_routes.append(self.parse_route_data(route))

        microservice_routes.sort(key=get_first_tag)

        for endpoint in microservice_routes:
            endpoint_name = list(endpoint.keys())[0]
            # if endpoint already exists, then update, else create a new one
            spec_endpoint = specification["paths"].get(endpoint_name)
            if spec_endpoint is None:
                spec_endpoint = specification["paths"][endpoint_name] = {}

            spec_endpoint.update(endpoint[endpoint_name])

        for endpoint in asab_routes:
            endpoint_name = list(endpoint.keys())[0]
            spec_endpoint = specification["paths"].get(endpoint_name)
            if spec_endpoint is None:
                spec_endpoint = specification["paths"][endpoint_name] = {}

            spec_endpoint.update(endpoint[endpoint_name])

        return specification

    def parse_route_data(self, route) -> dict:
        """
        Take a route (a single method of an endpoint) and return its description data.
        """
        path_parameters: list = extract_path_parameters(route)
        handler_name: str = get_handler_name(route)

        # Parse docstring description and yaml data
        docstring: str = route.handler.__doc__
        docstring_description: str = get_docstring_description(docstring)
        docstring_description += "\n\n**Handler:** `{}`".format(handler_name)
        docstring_yaml_dict: dict = self.get_docstring_yaml_dict(docstring)

        # Create route info dictionary
        route_info_data: dict = {
            "summary": docstring_description.split("\n")[0],
            "description": docstring_description,
            "responses": {"200": {"description": "Success"}},
            "parameters": [],
            "tags": [],
        }

        # Update it with parsed YAML and add query parameters
        if docstring_yaml_dict is not None:
            route_info_data.update(docstring_yaml_dict)
            for query_parameter in docstring_yaml_dict.get("parameters", []):
                if query_parameter.get("parameters"):
                    route_info_data["parameters"].append(query_parameter["parameters"])

        for path_parameter in path_parameters:
            route_info_data["parameters"].append(path_parameter)

        # Add default tag if not specified in docstring yaml
        if len(route_info_data["tags"]) == 0:
            # Use the default one
            if self.DefaultRouteTag == "class_name":
                route_info_data["tags"] = [get_class_name(route)]
            elif self.DefaultRouteTag == "module_name":
                route_info_data["tags"] = [get_module_name(route)]

        # Create the route dictionary
        route_path: str = self.get_route_path(route)
        method_name: str = route.method.lower()
        method_dict: dict = get_json_schema(route)
        method_dict.update(route_info_data)

        return {route_path: {method_name: method_dict}}

    def get_docstring_yaml_dict(
        self, docstring: typing.Optional[str]
    ) -> typing.Optional[dict]:
        """Take the docstring of a function and return additional data if they exist."""

        parsed_yaml_docstring_dict: typing.Optional[dict] = None

        if docstring is not None:
            docstring = inspect.cleandoc(docstring)
            dashes_index = docstring.find("\n---\n")
            if dashes_index >= 0:
                try:
                    parsed_yaml_docstring_dict = yaml.load(
                        docstring[dashes_index:], Loader=yaml.SafeLoader
                    )  # everything after --- goes to add_dict
                except yaml.YAMLError as e:
                    L.error(
                        "Failed to parse '{}' doc string {}".format(
                            self.App.__class__.__name__, e
                        )
                    )
        return parsed_yaml_docstring_dict

    def create_security_schemes(self) -> dict:
        """Create security schemes."""
        security_schemes_dict = {}
        if self.AuthorizationUrl and self.TokenUrl:
            security_schemes_dict = {
                "oAuth": {
                    "type": "oauth2",
                    "description": "",
                    "flows": {
                        "authorizationCode": {
                            "authorizationUrl": self.AuthorizationUrl,  # "http://localhost/seacat/api/openidconnect/authorize"
                            "tokenUrl": self.TokenUrl,  # "http://localhost/seacat/api/openidconnect/token"
                            "scopes": {
                                "openid": "Required Scope for OpenIDConnect!",
                            },
                        }
                    },
                }
            }
            if self.Scopes:
                for scope in self.Scopes.split(","):
                    security_schemes_dict["oAuth"]["flows"]["authorizationCode"][
                        "scopes"
                    ].update({"scope": "{} scope.".format(scope.strip().capitalize())})
        return security_schemes_dict

    def get_version_from_manifest(self) -> dict:
        """Get version from MANIFEST.json if exists."""
        if self.Manifest:
            version = self.Manifest["version"]
        else:
            version = "unknown"
        return version

    def get_route_path(self, route) -> str:
        """Take a route and return its path."""
        route_info = route.get_info()
        if "path" in route_info:
            path = route_info["path"]
        elif "formatter" in route_info:
            path = route_info["formatter"]
        else:
            L.warning("Cannot obtain path info from route", struct_data=self.route_info)
        return path

    @noauth
    # This is the web request handler
    async def doc(self, request):
        """
        Access the API documentation using a browser.
        ---
        tags: ['asab.doc']
        """

        swagger_js_url: str = "https://unpkg.com/swagger-ui-dist@4/swagger-ui-bundle.js"
        swagger_css_url: str = "https://unpkg.com/swagger-ui-dist@4/swagger-ui.css"

        doc_page = SWAGGER_DOC_PAGE.format(
            title=self.App.__class__.__name__,
            swagger_css_url=swagger_css_url,
            swagger_js_url=swagger_js_url,
            openapi_url="./asab/v1/openapi",
        )

        return aiohttp.web.Response(text=doc_page, content_type="text/html")

    @noauth
    async def oauth2_redirect(self, request):
        """
        Required for the authorization to work.
        ---
        tags: ['asab.doc']
        """

        return aiohttp.web.Response(text=SWAGGER_OAUTH_PAGE, content_type="text/html")

    @noauth
    async def openapi(self, request):
        """
        Download OpenAPI (version 3) API documentation (aka Swagger) in YAML.
        ---
        tags: ['asab.doc']
        externalDocs:
        description: OpenAPI Specification
        url: https://swagger.io/specification/

        """
        return aiohttp.web.Response(
            text=(yaml.dump(self.build_swagger_documentation(), sort_keys=False)),
            content_type="text/yaml",
        )


def get_docstring_description(docstring: typing.Optional[str]) -> str:
    """Take the docstring of a function and parse it into description. Omit everything that comes after '---'."""
    if docstring is not None:
        docstring = inspect.cleandoc(docstring)
        dashes_index = docstring.find(
            "\n---\n"
        )  # find the index of the first three dashes

        # everything before --- goes to description
        if dashes_index >= 0:
            description = docstring[:dashes_index]
        else:
            description = docstring
    else:
        description = ""
    return description


def extract_path_parameters(route) -> list:
    """Take a single route and return its parameters."""
    parameters: list = []
    route_info = route.get_info()
    if "formatter" in route_info:
        path = route_info["formatter"]
        for params in re.findall(r"\{[^\}]+\}", path):
            parameters.append(
                {
                    "in": "path",
                    "name": params[1:-1],
                    "required": True,
                }
            )
    return parameters


def get_handler_name(route) -> str:
    if inspect.ismethod(route.handler):
        handler_name = "{}.{}()".format(
            route.handler.__self__.__class__.__name__, route.handler.__name__
        )
    else:
        handler_name = "{}()".format(route.handler.__qualname__)
    return handler_name


def get_class_name(route) -> str:
    if inspect.ismethod(route.handler):
        class_name = str(route.handler.__self__.__class__.__name__)
    else:
        class_name = str(route.handler.__qualname__.split(".")[0])
    return class_name


def get_module_name(route) -> str:
    return str(route.handler.__module__)


def get_json_schema(route) -> dict:
    method_dict = {}
    try:
        json_schema = route.handler.__getattribute__("json_schema")
        method_dict["requestBody"] = {
            "content": {"application/json": {"schema": json_schema}},
        }
    except AttributeError:
        pass
    return method_dict


def get_first_tag(route_data: dict) -> str:
    """Get tag from route data. Used for sorting tags alphabetically."""
    for endpoint in route_data.values():
        for method in endpoint.values():
            if method.get("tags"):
                return method.get("tags")[0].lower()



from .test_declarative_datetime import *
from .test_declarative_config import *
from .test_declarative_join import *
from .test_declarative_dict import *
from .test_declarative_regex import *
from .test_declarative_when import *
from .test_declarative_dict_parse import *
from .test_declarative_and import *
from .test_declarative_add import *
from .test_declarative_time import *
from .test_declarative_nested_expression import *



from bspump.http.web.components.base_field import BaseField
from bspump.http.web.template_env import env


class FieldSet(BaseField):
    def __init__(
        self, name, fields=None, fieldset_intro="", display="", required=True, **kwargs
    ):
        super().__init__(name, **kwargs)
        su = super()
        self.fields = fields
        if fields is None:
            self.fields = []
        su.__init__(name, required=required)
        self.display = display if display else name
        self.default = {}
        self.fieldset_intro = fieldset_intro
        self.prefix = f"fieldset___{self.name}___"

    def set_subfield_names(self):
        for field in self.fields:
            field.field_name = f"{self.prefix}{field.name}"

    def html(self, defaults={}):
        self.set_subfield_names()
        self.set_subfield_names()
        fields_html = [
            field.html(defaults.get(field.name, field.default)) for field in self.fields
        ]
        template = env.get_template("fieldset.html")
        return template.render(
            display=self.display, fieldset_intro=self.fieldset_intro, fields=fields_html
        )

    def restructure_data(self, dfrom, dto):
        dto[self.name] = {}
        for field in self.fields:
            field.restructure_data(dfrom, dto[self.name])

    def get_params(self) -> dict:
        out = {}
        for field in self.fields:
            out[field.name] = field.get_params()
        return out

    def clean(self, data, request=None):
        for field in self.fields:
            field.clean(data[self.name], request)



import asyncio
import collections.abc
import json
import logging
from typing import Optional

import bspump.asab as asab

from .lookupprovider import LookupProviderABC

###

L = logging.getLogger(__name__)

###


class Lookup(asab.Configurable):
    """
    Description:

    |

    :return:
    """

    ConfigDefaults = {
        "source_url": "",  # Specifies complete url to source file
        # zk://zookeeper1:2181/base/path/to/config.yaml
        # zk:///base/path/to/config.yaml  ==  zk:/base/path/to/config.yaml
        # zk:///./path/to/config.yaml  ==  zk:/./path/to/config.yaml
        # http://localhost:8080/path/to/config.yaml
        # file:/root/path/to/config.yaml  ==  /root/path/to/config.yaml
        # Backwards compatibility
        "master_url": "http://localhost:8080",
        "master_url_endpoint": "/bspump/v1/lookup/",
        "master_lookup_id": "",  # If not empty, it specify the lookup id that will be used for loading from master
    }

    def __init__(self, app, id=None, config=None, lazy=False):
        """
        Description:

        """
        _id = id if id is not None else self.__class__.__name__
        super().__init__("lookup:{}".format(_id), config=config)
        self.Id = _id
        self.App = app
        self.Loop = app.Loop
        self.Lazy = lazy

        self.PubSub = asab.PubSub(app)

        self.MasterURL = None
        self.Provider: Optional[LookupProviderABC] = None

        url = self.Config.get("source_url", "").strip()
        if len(url) == 0:
            # Construct URL from the old "master_" params
            server = self.Config.get("master_url", "").strip()
            if len(server) == 0:
                L.error(
                    "Neither `source_url` nor `master_url` specified in lookup {} config.".format(
                        self.Id
                    )
                )
                return
            if not server.startswith("http:"):
                server = "http://{}".format(server)
            server = server.strip("/")
            master_url_endpoint = self.Config["master_url_endpoint"].strip("/")
            master_lookup_id = self.Config["master_lookup_id"]
            if master_lookup_id == "":
                master_lookup_id = self.Id
            url = "{}/{}/{}".format(server, master_url_endpoint, master_lookup_id)
        self._create_provider(url)

    def __getitem__(self, key):
        raise NotImplementedError(
            "Lookup '{}' __getitem__() method not implemented".format(self.Id)
        )

    def __iter__(self):
        raise NotImplementedError(
            "Lookup '{}' __iter__() method not implemented".format(self.Id)
        )

    def __len__(self):
        raise NotImplementedError(
            "Lookup '{}' __len__() method not implemented".format(self.Id)
        )

    def __contains__(self, item):
        raise NotImplementedError(
            "Lookup '{}' __contains__() method not implemented".format(self.Id)
        )

    def _create_provider(self, path: str):
        """
        Description:

        :return:
        """
        if path.startswith("zk:"):
            from bspump.zookeeper import ZooKeeperBatchLookupProvider

            self.Provider = ZooKeeperBatchLookupProvider(self, path)
            self.MasterURL = path
        elif path.startswith("http:") or path.startswith("https:"):
            from bspump.http import HTTPBatchLookupProvider

            config = {}
            if "use_cache" in self.Config:
                config["use_cache"] = self.Config.getboolean("use_cache")
            if "cache_dir" in self.Config:
                config["cache_dir"] = self.Config.get("source_url", None)
            self.Provider = HTTPBatchLookupProvider(self, path, config=config)
            self.MasterURL = path
        else:
            from bspump.file import FileBatchLookupProvider

            # Local file source -> lookup is considered master
            self.Provider = FileBatchLookupProvider(self, path)
            self.MasterURL = None

    def time(self):
        """
        Description:

        :return: time

        |

        """
        return self.App.time()

    def ensure_future_update(self, loop):
        """
        Description:

        :return:

        |

        """
        return asyncio.ensure_future(self._do_update())

    async def _do_update(self):
        """
        Description:

        :return:
        """
        updated = await self.load()
        if updated:
            L.warning(f"{self.Id} bspump.Lookup.changed!")
            self.PubSub.publish("bspump.Lookup.changed!")

    async def load(self) -> bool:
        """
        Description:

        """
        data = await self.Provider.load()
        if data is None or data is False:
            L.warning("No data loaded from {}.".format(self.Provider.Id))
            return False
        self.deserialize(data)
        return True

    def serialize(self):
        """
        Description:

        """
        raise NotImplementedError(
            "Lookup '{}' serialize() method not implemented".format(self.Id)
        )

    def deserialize(self, data):
        """
        Description:

        |

        """
        raise NotImplementedError(
            "Lookup '{}' deserialize() method not implemented".format(self.Id)
        )

    def rest_get(self):
        """
        Description:

        :return:
        """
        response = {"Id": self.Id}
        if self.Provider.ETag is not None:
            response["ETag"] = self.Provider.ETag
        if self.MasterURL is not None:
            response["MasterURL"] = self.MasterURL
        return response

    def is_master(self):
        """
        Description:

        :return: ??

        |

        """
        return self.MasterURL is None


class MappingLookup(Lookup, collections.abc.Mapping):
    """
    Description:

    |

    """

    pass


class AsyncLookupMixin(Lookup):
    """
    Description:

    """

    async def get(self, key):
        raise NotImplementedError()


class DictionaryLookup(MappingLookup):
    """
    Description:

    """

    def __init__(self, app, id=None, config=None, lazy=False):
        """
        Description:

        |

        """

        self.Dictionary = {}
        super().__init__(app, id, config=config, lazy=lazy)

    def __getitem__(self, key):
        return self.Dictionary.__getitem__(key)

    def __iter__(self):
        return self.Dictionary.__iter__()

    def __len__(self):
        return self.Dictionary.__len__()

    def serialize(self):
        """
        Description:

        :return: json data

        |
        """
        return (json.dumps(self.Dictionary)).encode("utf-8")

    def deserialize(self, data):
        """
        Description:

        |

        """
        self.Dictionary.update(json.loads(data.decode("utf-8")))

    # REST

    def rest_get(self):
        """
        Description:

        :return: rest

        |
        """
        rest = super().rest_get()
        rest["Dictionary"] = self.Dictionary
        return rest

    def set(self, dictionary: dict):
        """
        Description:

        |

        """
        if self.is_master() is False:
            L.warning("'master_url' provided, set() method can not be used")

        self.Dictionary.clear()
        self.Dictionary.update(dictionary)



import asyncio
import logging
import os

import time

from .globscan import iter_files_glob
from ..abc.source import TriggerSource


L = logging.getLogger(__file__)


class FileABCSource(TriggerSource):
    """
    Description:

    """

    ConfigDefaults = {
        "path": "",
        "mode": "rb",
        "newline": os.linesep,
        "post": "move",  # one of 'delete', 'noop' and 'move'
        "exclude": "",  # glob of filenames that should be excluded (has precedence over 'include')
        "include": "",  # glob of filenames that should be included
        "encoding": "",
        "move_destination": "",  # destination folder for 'move'. Make sure it's outside of the glob search
        "lines_per_event": 10000,  # the number of lines after which the read method enters the idle state to allow other operations to perform their tasks
        "event_idle_time": 0.01,  # the time for which the read method enters the idle state (see above)
        "files_per_cycle": 1,
    }

    def __init__(self, app, pipeline, id=None, config=None):
        """
        Description:

        **Parameters**

        app : Application
                Name of the Application.

        pipeline : Pipeline
                Name of the Pipeline.

        id : ID, default = None
                ID

        config : JSON, default = None
                Configuration file with additional information.
                path : str (required)
                        Path to the file. Can be a glob pattern to select multiple files.
                mode : str, default = 'rb'
                        Mode in which the file is opened.
                newline : str, default = os.linesep
                        Newline character.
                post : str, default = 'move'
                        One of 'delete', 'noop' and 'move'.
                exclude : str, default = ''
                        Glob of filenames that should be excluded (has precedence over 'include').
                include : str, default = ''
                        Glob of filenames that should be included.
                encoding : str, default = ''
                        Encoding of the file.
                move_destination : str, default = ''
                        Destination folder for 'move'. Make sure it's outside of the glob search.
                lines_per_event : int, default = 10000
                        The number of lines after which the read method enters the idle state to allow other operations to perform their tasks.
                event_idle_time : float, default = 0.01
                        The time for which the read method enters the idle state (see above).
                files_per_cycle : int, default = 1
                        The number of files that are processed in one cycle.
        """
        super().__init__(app, pipeline, id=id, config=config)
        self.path = self.Config["path"]
        self.mode = self.Config["mode"]
        self.newline = self.Config["newline"]
        self.post = self.Config["post"]
        if self.post not in ["delete", "noop", "move"]:
            L.warning(
                "Incorrect/unknown 'post' configuration value '{}' - defaulting to 'move'".format(
                    self.post
                )
            )
            self.post = "move"
        self.include = self.Config["include"]
        self.exclude = self.Config["exclude"]
        conf_encoding = self.Config["encoding"]
        conf_encoding = self.Config["encoding"]
        self.files_per_cycle = self.Config["files_per_cycle"]
        if type(self.files_per_cycle) is not int:
            try:
                self.files_per_cycle = int(self.files_per_cycle)
            except ValueError:
                L.error(
                    "Incorrect 'files_per_cycle' configuration value '{}'".format(
                        self.files_per_cycle
                    )
                )
        self.encoding = conf_encoding if len(conf_encoding) > 0 else None

        self.MoveDestination = self.Config["move_destination"]

        if self.MoveDestination != "":
            if (self.post == "move") and (not os.path.isdir(self.MoveDestination)):
                os.makedirs(self.MoveDestination)
        else:
            self.MoveDestination = None

        metrics_service = app.get_service("asab.MetricsService")
        self.Gauge = metrics_service.create_gauge(
            "file_count",
            tags={
                "pipeline": pipeline.Id,
            },
            init_values={
                "processed": 0,
                "failed": 0,
                "locked": 0,
                "unprocessed": 0,
                "all_files": 0,
                "scan_time": 0.0,
            },
        )

        self.Loop = app.Loop
        self.ProactorService = app.get_service("asab.ProactorService")

        self.LinesCounter = 0
        self.LinesPerEvent = int(self.Config["lines_per_event"])
        self.EventIdleTime = float(self.Config["event_idle_time"])

    async def cycle(self):
        """
        Cycles through files that match the glob pattern.
        """
        filename = None

        start_time = time.time()
        filenames = []
        for path in self.path.split(os.pathsep):
            these_filenames = list(
                await self.ProactorService.execute(
                    iter_files_glob,
                    path,
                    self.Gauge,
                    self.Loop,
                    self.exclude,
                    self.include,
                )
            )
            filenames.extend(these_filenames)
            if len(filenames) >= self.files_per_cycle:
                break
        end_time = time.time()
        self.Gauge.set("scan_time", end_time - start_time)

        if len(filenames) == 0:
            self.Pipeline.PubSub.publish("bspump.file_source.no_files!")
            return  # No file to read

        await self.Pipeline.ready()
        for filename in filenames[: self.files_per_cycle]:
            await self.lock_and_read_file(filename)

    async def lock_and_read_file(self, filename):
        # Lock the file
        L.debug("Locking file '{}'".format(filename))
        locked_filename = filename + "-locked"
        try:
            os.rename(filename, locked_filename)
        except FileNotFoundError:
            return
        except (OSError, PermissionError):  # OSError - UNIX, PermissionError - Windows
            L.exception(
                "Error when locking the file '{}'  - will try again".format(filename)
            )
            return
        except BaseException as e:
            L.exception("Error when locking the file '{}'".format(filename))
            self.Pipeline.set_error(None, None, e)
            return

        try:
            if filename.endswith(".gz"):
                import gzip

                f = gzip.open(locked_filename, self.mode, encoding=self.encoding)

            elif filename.endswith(".bz2"):
                import bz2

                f = bz2.open(locked_filename, self.mode, encoding=self.encoding)

            elif filename.endswith(".xz") or filename.endswith(".lzma"):
                import lzma

                f = lzma.open(locked_filename, self.mode, encoding=self.encoding)

            else:
                if "b" in self.mode:  # Binary mode doesn't take a newline argument
                    self.newline = None
                f = open(
                    locked_filename,
                    self.mode,
                    newline=self.newline,
                    encoding=self.encoding,
                )

        except (OSError, PermissionError):  # OSError - UNIX, PermissionError - Windows
            L.exception(
                "Error when opening the file '{}' - will try again".format(filename)
            )
            return
        except BaseException as e:
            L.exception("Error when opening the file '{}'".format(filename))
            self.Pipeline.set_error(None, None, e)
            return

        L.debug("Processing file '{}'".format(filename))

        try:
            await self.read(filename, f)
        except Exception:
            try:
                if self.post == "noop":
                    # When we should stop, rename file back to original
                    os.rename(locked_filename, filename)
                else:
                    # Otherwise rename to ...-failed and continue processing
                    os.rename(locked_filename, filename + "-failed")
            except BaseException:
                L.exception(
                    "Error when renaming the file '{}'  - will try again".format(
                        filename
                    )
                )
            return
        finally:
            f.close()

        L.debug("File '{}' processed {}".format(filename, "succefully"))

        # Finalize
        try:
            if self.post == "delete":
                os.unlink(locked_filename)
            elif self.post == "noop":
                os.rename(locked_filename, filename)
            else:
                if self.MoveDestination is not None:
                    file_from = os.path.abspath(locked_filename)
                    base = os.path.basename(filename)
                    file_to = os.path.abspath(
                        os.path.join(self.MoveDestination, base + "-processed")
                    )
                else:
                    file_from = locked_filename
                    file_to = filename + "-processed"

                os.rename(file_from, file_to)
        except (OSError, PermissionError):  # OSError - UNIX, PermissionError - Windows
            L.exception(
                "Error when finalizing the file '{}' - will try again".format(filename)
            )
            return
        except BaseException as e:
            L.exception("Error when finalizing the file '{}'".format(filename))
            self.Pipeline.set_error(None, None, e)
            return

    async def simulate_event(self):
        """
        The simulate_event method should be called in read method after a file line has been processed.

        It ensures that all other asynchronous events receive enough time to perform their tasks.
        Otherwise, the application loop is blocked by a file reader and no other activity makes a progress.

        """
        self.LinesCounter += 1
        if self.LinesCounter >= self.LinesPerEvent:
            await asyncio.sleep(self.EventIdleTime)
            self.LinesCounter = 0

    async def read(self, filename, f):
        """
        Description: Override this method to implement your File Source.
        `f` is an opened file object.

        **Parameters**

        filename : file
                Name of the file.

        f :

        """
        raise NotImplementedError()


try:
    import pytest

    @staticmethod
    @pytest.mark.asyncio
    async def test_file_abc_source():
        await run_file_abc_source(
            "move",
            ["1-processed", "2-processed", "3", "4-locked", "5-failed", "6-processed"],
        )
        await run_file_abc_source(
            "noop", ["1", "2", "3", "4-locked", "5-failed", "6-processed"]
        )
        await run_file_abc_source(
            "delete", ["3", "4-locked", "5-failed", "6-processed"]
        )

    async def run_file_abc_source(post, expected_resultant_files):
        read_files = []

        class TestFileABCSource(FileABCSource):
            async def read(self, filename, f):
                read_files.append(filename)

        # Copy test files to temp dir
        import shutil
        import tempfile

        temp_dir = tempfile.mkdtemp()
        # copy globscan test data ("bspump/test-data/globscan/") to temp dir
        shutil.copytree(
            os.path.join(os.path.dirname(__file__), "..", "test-data", "globscan"),
            os.path.join(temp_dir, "globscan"),
        )

        class FakePipeline:
            def __init__(self):
                self.Id = "FakePipeline"
                self.Loop = asyncio.get_event_loop()

                class FakePubSub:
                    def publish(self, *args, **kwargs):
                        pass

                self.PubSub = FakePubSub()

            async def ready(self):
                pass

            def get_service(self, *args, **kwargs):
                class FakeService:
                    def create_gauge(self, *args, **kwargs):
                        class FakeGauge:
                            def set(self, *args, **kwargs):
                                pass

                        return FakeGauge()

                    async def execute(self, fn, *args, **kwargs):
                        return fn(*args, **kwargs)

                return FakeService()

        testfilesource = TestFileABCSource(
            app=FakePipeline(),
            config={
                "path": os.path.join(temp_dir, "globscan", "*"),
                "files_per_cycle": "2",
                "post": post,
            },
            pipeline=FakePipeline(),
        )
        await testfilesource.cycle()
        assert len(read_files) == 2
        assert read_files[0] == os.path.join(temp_dir, "globscan", "1")
        assert read_files[1] == os.path.join(temp_dir, "globscan", "2")
        # Make sure that the files have been renamed as being processed
        # Get file listing and compare for easy debugging
        assert sorted(os.listdir(os.path.join(temp_dir, "globscan"))) == sorted(
            expected_resultant_files
        )

except ImportError:
    pass
